"""
Base class for operations.
"""
from typing import Any, List, Optional, Union
from relevanceai.client.helpers import (
    Credentials,
    process_token,
)


class BaseOps:
    """
    Base class for operations
    """

    def __init__(self, *args, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def init(self, *args, **kwargs):
        return self(*args, **kwargs)

    @classmethod
    def from_credentials(self, credentials: Credentials, *args, **kw):
        return self(credentials=credentials)

    @classmethod
    def from_token(self, token: str, *args, **kw):
        """
        If this is from a token, then we use this
        """
        credentials = process_token(token)
        return self(credentials=credentials, *args, **kw)

    @classmethod
    def from_client(self, client, *args, **kwargs):
        credentials = client.credentials
        return self(
            credentials=credentials,
            *args,
            **kwargs,
        )

    @classmethod
    def from_dataset(
        self, dataset: Any, alias: str, vector_fields: List[str], *args, **kwargs
    ):
        dataset_id = dataset.dataset_id
        credentials = dataset.credentials
        return self(
            credentials=credentials,
            dataset_id=dataset_id,
            alias=alias,
            vector_fields=vector_fields,
            vector_field=vector_fields[0],
            *args,
            **kwargs,
        )

    def _get_dataset_id(self, dataset: Union[str, Any]):
        from relevanceai.dataset import Dataset

        if isinstance(dataset, str):
            return dataset
        elif isinstance(dataset, Dataset):
            return dataset.dataset_id

    def _check_vector_fields(self, vector_fields: List[str]):
        if vector_fields is None:
            raise ValueError("Vector fields cannot be None. Please set vector_fields=")
        if len(vector_fields) == 0:
            raise ValueError(
                "Please select at least 1 vector field"
            )  # we can add a optional behaviour to use all vectors here.
        if isinstance(vector_fields, str):
            return [vector_fields]
        else:
            new_vector_fields = []
            for f in vector_fields:
                if not f.endswith(
                    "_vector_"
                ):  # this could support chunkvector in the future.
                    new_vector_fields.append(
                        f + "_vector_"
                    )  # this will modify input, but assumes this is a mistake that the user is making
                else:
                    new_vector_fields.append(f)
            return new_vector_fields

    def _create_alias(self, alias, vector_fields, tag="default"):
        vector_suffix = f"_{tag}_vector_" if tag else "_vector_"
        if alias is None:
            if vector_fields is not None:
                if len(vector_fields) > 0:
                    new_alias = (
                        "-".join([f.replace("_vector_", "") for f in vector_fields])
                        + vector_suffix
                    )
                elif len(vector_fields) == 1:
                    new_alias = vector_fields[0] + vector_suffix
                else:
                    new_alias = vector_suffix
            else:
                new_alias = vector_suffix
            return new_alias
        else:
            return alias if alias.endswith("_vector_") else alias + vector_suffix



from relevanceai.operations_new.cluster.batch.models.sklearn.base import (
    SklearnBatchClusterBase,
)


class MiniBatchKMeans(SklearnBatchClusterBase):
    @property
    def name(self):
        return "minibatchkmeans"



from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base
from typing import List


class FieldChildrenClient(_Base):
    """All dataset-related functions"""

    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    def list(self, dataset_id: str):
        """
        Returns the schema of a dataset. Refer to datasets.create for different field types available in a Relevance schema.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/field_children/list",
            method="POST",
        )

    def delete_field_children(self, dataset_id: str, fieldchildren_id: str):
        """
        Returns the schema of a dataset. Refer to datasets.create for different field types available in a Relevance schema.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/field_children/{fieldchildren_id}/delete",
            method="POST",
        )

    delete = delete_field_children

    def update_field_children(
        self,
        dataset_id: str,
        fieldchildren_id: str,
        field: str,
        field_children: List,
        metadata: dict = None,
    ):
        if metadata is None:
            metadata = {}
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/field_children/{fieldchildren_id}/update",
            method="POST",
            parameters={
                "field": field,
                "field_children": field_children,
                "metadata": metadata,
            },
        )

    # make compatible with endpoint
    update = update_field_children



"""Testing code for batch inserting
"""
import pytest

from typing import Dict, List

from relevanceai import Client

from relevanceai.dataset import Dataset

from tests.globals.constants import generate_dataset_id

from tests.unit.test_api.helpers import *


class TestInsert:
    test_dataset_id = generate_dataset_id()

    def test_batch_insert(
        self,
        test_dataset: Dataset,
        vector_documents: List[Dict],
    ):
        results = test_dataset.insert_documents(vector_documents)
        assert len(results["failed_documents"]) == 0

    def test_assorted_nested_upload(
        self,
        test_dataset: Dataset,
        assorted_nested_documents: List[Dict],
    ):
        results = test_dataset.insert_documents(assorted_nested_documents)
        assert len(results["failed_documents"]) == 0


class TestInsertImages:
    def setup(self):
        from pathlib import Path
        from uuid import uuid4

        self.filename = "lovelace.jpg"

        self.directory = Path(str(uuid4()))
        self.directory.mkdir()
        with open(self.filename, "wb") as f:
            f.write(b"ghuewiogahweuaioghweqrofleuwaiolfheaswufg9oeawhfgaeuw")

    @pytest.mark.skip(reason="need to fix image folder")
    def test_insert_media_folder(self, test_client: Client):
        self.ds = test_client.Dataset(generate_dataset_id())
        results = self.ds.insert_media_folder(
            field="images",
            path=self.directory,
            recurse=False,  # No subdirectories exist anyway
        )
        assert results is None

    def teardown(self):
        self.ds.delete()

        try:
            import os

            os.remove(self.directory / self.filename)
        except FileNotFoundError:
            pass

        self.directory.rmdir()



from relevanceai.utils import MissingPackageError
from relevanceai.operations_new.transform_base import TransformBase


class CountTextTransform(TransformBase):
    _imported_nltk = False

    def __init__(
        self,
        text_fields: list,
        include_char_count: bool = True,
        include_word_count: bool = True,
        include_sentence_count: bool = False,
        output_fields: list = None,
    ):
        self.text_fields = text_fields
        self.include_char_count = include_char_count
        self.include_word_count = include_word_count
        self.include_sentence_count = include_sentence_count
        self.output_fields = output_fields

    def count_characters(self, text):
        return len(text)

    def count_words(self, text):
        return len(text.split())

    def count_sentences(self, text):
        try:
            from sentence_splitter import split_text_into_sentences
        except ModuleNotFoundError:
            raise MissingPackageError("sentence-splitter")
        sentences = split_text_into_sentences(
            text=text,
            language="en",
        )
        return len(sentences)

    def count_text_document(self, document):
        try:
            output_doc = {"_id": document["_id"]}
            for i, t in enumerate(self.text_fields):
                text = self.get_field(t, document)
                if self.include_char_count:
                    self.set_field(
                        "_count_.char." + t, output_doc, self.count_characters(text)
                    )
                if self.include_word_count:
                    self.set_field(
                        "_count_.word." + t, output_doc, self.count_words(text)
                    )
                if self.include_sentence_count:
                    self.set_field(
                        "_count_.sentence." + t, output_doc, self.count_sentences(text)
                    )
        except:
            pass
        return output_doc

    def transform(self, documents: list):
        new_docs = [self.count_text_document(d) for d in documents]
        return new_docs



"""
Labelling performs a vector search on the labels and fetches the closest
max_number_of_labels.

"""
from copy import deepcopy
from typing import Any, Dict, List

from relevanceai.operations_new.transform_base import TransformBase


class LabelTransform(TransformBase):
    def __init__(
        self,
        vector_field: str,
        label_documents: list,
        expanded: bool = True,
        max_number_of_labels: int = 1,
        similarity_metric: str = "cosine",
        similarity_threshold: float = 0.1,
        label_field="label",
        label_vector_field="label_vector_",
        output_field: str = "_label_",
        **kwargs,
    ):
        self.vector_field = vector_field
        self.expanded = expanded
        self.max_number_of_labels = max_number_of_labels
        self.similarity_metric = similarity_metric
        self.similarity_threshold = similarity_threshold
        self.label_field = label_field
        self.label_vector_field = label_vector_field
        self.label_documents = label_documents
        self.vector_fields = [vector_field]
        self.output_field = output_field

        for k, v in kwargs.items():
            setattr(self, k, v)

    def transform(  # type: ignore
        self,
        documents,
    ) -> List[Dict[str, Any]]:

        """Get all vectors, search across

        Parameters
        ----------
        documents
            the documents to be labeled
        label_documents
            The documents that contain the labels.

        Example
        -------
        .. code-block::

            ds = client.Dataset(...)
            # label an entire dataset
            ds.label(
                vector_field="sample_1_vector_",
                label_documents=[
                    {
                        "label": "value",
                        "price": 0.3,
                        "label_vector_": [1, 1, 1]
                    },
                    {
                        "label": "value-2",
                        "label_vector_": [2, 1, 1]
                    },
                ],
                expanded=True # stored as dict or list
            )

        If missing "label", returns Error - labels missing `label` field
        writes loop to set `label` field

        If you want all values in a label document plus similarity, you need to set
        expanded=True

        Returns
        -------
            A list of dictionaries.

        """

        # Get all vectors
        vectors = self.get_field_across_documents(self.vector_field, documents)

        # TODO switch this to multiprocessing
        from relevanceai.utils import multiprocess_list

        # label_docs = multiprocess_list(self.get_label_document, documents)
        label_docs = [
            self.get_label_document(document) for i, document in enumerate(documents)
        ]

        return label_docs

    def get_label_document(self, document, *args, **kwargs):
        labels = self._get_nearest_labels(
            vector=self.get_field(self.vector_field, document),
            label_documents=self.label_documents,
        )
        doc: dict = {"_id": document["_id"]}
        self.set_field(self.output_field, doc, labels)
        return doc

    @property
    def name(self):
        return "labelling"

    def _get_nearest_labels(
        self,
        vector: List[float],
        label_documents: List[Dict[str, Any]],
    ):
        """It takes a vector and a list of documents, and returns a list of labels

        Parameters
        ----------
        vector : List[float]
            List[float]
        label_documents : List[Dict[str, Any]]
            List[Dict[str, Any]]

        Returns
        -------
            The return value is a list of labels.

        """
        # perform cosine similarity
        if self.similarity_metric == "cosine":
            labels = self.cosine_similarity(
                query_vector=vector,
                vector_field=self.label_vector_field,
                documents=label_documents,
                max_number_of_labels=self.max_number_of_labels,
                similarity_threshold=self.similarity_threshold,
            )
        else:
            raise ValueError(
                "Only cosine similarity metric is supported at the moment."
            )

        # for the label vectors
        if self.expanded:
            return labels
        else:
            # get a list of labels
            return self.get_field_across_documents(self.label_field, labels)

    def cosine_similarity(
        self,
        query_vector,
        vector_field,
        documents,
        reverse=True,
        score_field: str = "_label_score",
        max_number_of_labels: int = 1,
        similarity_threshold: float = 0,
    ):
        """It takes a query vector, a vector field, a list of documents, and a few other parameters, and
        returns a list of documents sorted by their cosine similarity to the query vector

        Parameters
        ----------
        query_vector
            the vector you want to compare against
        vector_field
            the field in the documents that contains the vector
        documents
            list of documents
        reverse, optional
            True/False
        score_field : str, optional
            str = "_label_score"
        max_number_of_labels : int, optional
            int = 1,
        similarity_threshold : float, optional
            float = 0,

        Returns
        -------
            A list of dictionaries.

        """
        from scipy.spatial import distance

        sort_key = [
            1 - distance.cosine(i, query_vector)
            for i in self.get_field_across_documents(vector_field, documents)
        ]
        self.set_field_across_documents(score_field, sort_key, documents)
        labels = sorted(documents, reverse=reverse, key=lambda x: x[score_field])
        labels = [l for l in labels if l[score_field] > similarity_threshold]
        counter = 0
        new_labels: List[dict] = []
        for label in labels:
            label_text = self.get_field(self.label_field, label)
            label_texts = self.get_field_across_documents(self.label_field, new_labels)
            if label_text not in label_texts:
                new_labels.append(deepcopy(label))
                counter += 1
                if counter >= max_number_of_labels:
                    [l.pop(vector_field) for l in new_labels]
                    return new_labels
        # new_labels = deepcopy(labels)
        # remove labels from labels
        [l.pop(vector_field) for l in new_labels]
        return new_labels

    def get_operation_metadata(self) -> Dict[str, Any]:
        return dict(
            operation="label",
            values=str(
                {
                    "vector_fields": self.vector_fields,
                    "expanded": self.expanded,
                    "max_number_of_labels": self.max_number_of_labels,
                    "similarity_metric": self.similarity_metric,
                    "similarity_threshold": self.similarity_threshold,
                    "label_field": self.label_field,
                    "label_vector_field": self.label_vector_field,
                }
            ),
        )



from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.vectorize.transform import VectorizeTransform


class VectorizeOps(VectorizeTransform, OperationAPIBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from collections.abc import MutableSequence

from typing import Dict, List, Any

from pandas import isna

from .chunk_doc_utils import ChunkDocUtils

try:
    from IPython.display import display
except ModuleNotFoundError:
    pass


class DocUtils(ChunkDocUtils):
    """Class for all document utilities.
    Primarily should be used as a mixin for future functions
    but can be a standalone.
    # TODO: Extend to Chunk Doc Reading and Chunk Doc Writing
    """

    pass


class Document(DocUtils):
    def __init__(self, document: Dict):
        """It takes a dictionary and returns a Document object

        Parameters
        ----------
        document : Dict
            Dict

        """
        super().__init__()

        for key, value in document.items():
            if isinstance(value, dict):
                document[key] = Document(value)

        self.data = document

    """> The `_ipython_display_` function is called by the IPython notebook when it wants to display an
    object

    """

    def _ipython_display_(self):
        display(self.json())

    def __repr__(self):
        """It returns a string representation of the object

        Returns
        -------
            The json() method is being called on the object.

        """
        return str(self.json())

    def __len__(self):
        """It returns the length of the JSON object

        Returns
        -------
            The length of the json object

        """
        return len(self.json())

    def __eq__(self, other):
        """If the keys and values of the two dictionaries are the same, then the dictionaries are the same

        Parameters
        ----------
        other
            The other object to compare to.

        """
        try:
            for key1, key2, value1, value2 in zip(
                self.keys(),
                other.keys(),
                self.values(),
                other.values(),
            ):
                if value1 != value2:
                    return False
            return True
        except:
            return False

    def __getattr__(self, name):
        """If the attribute name starts with two underscores, raise an AttributeError. Otherwise, if the
        attribute name is in the data dictionary, return the value from the dictionary. Otherwise, raise
        an AttributeError

        Parameters
        ----------
        name
            The name of the parameter.

        Returns
        -------
            A class object.

        """
        if name.startswith("__"):
            raise AttributeError
        if hasattr(self, "data"):
            if name in self.data:
                return self.data[name]
        raise AttributeError

    def __getitem__(self, key: str) -> Any:
        """It takes a string, splits it into a list of strings, and then uses the first string to get the
        first value, the second string to get the second value, and so on

        Parameters
        ----------
        key : str
            str

        Returns
        -------
            The value of the key.

        """
        levels = key.split(".")

        value = self.data
        for level in levels:
            value = value.__getitem__(level)

        return value

    """> The `__iter__` function returns an iterator object that iterates over the keys of the dictionary

    Returns
    -------
        The keys of the dictionary.

    """

    def __iter__(self):
        return iter(self.keys())

    def __setitem__(self, key: str, value: Any) -> None:
        """It takes a key and a value, splits the key into a list of keys, and then calls the setitem
        function with the list of keys and the value

        Parameters
        ----------
        key : str
            str
        value : Any
            The value to set.

        """
        self.setitem(self, key.split("."), value)

    def setitem(self, obj, keys: List[str], value: Any) -> None:
        """> It takes a list of keys, and a value, and sets the value at the end of the list of keys

        Parameters
        ----------
        obj
            The object to set the value on.
        keys : List[str]
            List[str]
        value : Any
            The value to set.

        """
        for key in keys[:-1]:
            obj = obj.data.setdefault(key, {})
        obj.data[keys[-1]] = value

    def json(self):
        """If the value is a DataFrame, then call the json() function on it. Otherwise, if the value is a
        NaN, then return None. Otherwise, return the value

        Returns
        -------
            A dictionary of dictionaries.

        """
        document = {}
        for key, value in self.data.items():
            if isinstance(value, self.__class__):
                document[key] = value.json()
            else:
                try:
                    if isna(value):
                        document[key] = None
                    else:
                        document[key] = value
                except:
                    document[key] = value
        return document

    def keys(self, parent=None):
        """It takes a dictionary, and returns a list of all the keys in the dictionary, including nested
        keys

        Parameters
        ----------
        parent
            The parent key of the current dictionary.

        Returns
        -------
            A list of keys

        """
        keys = {}

        for key, value in self.data.items():
            if isinstance(value, self.__class__):
                if parent is not None:
                    subparent = f"{parent}.{key}"
                else:
                    subparent = key
                keys.update({key: None for key in value.keys(parent=subparent)})
                keys.update({subparent: None})
            else:
                if parent is None:
                    keys[key] = None
                else:
                    keys[f"{parent}.{key}"] = None

        return keys.keys()

    def get(self, key, value=None):
        """If the key is in the dictionary, return the value, otherwise return the default value

        Parameters
        ----------
        key
            The key to be searched in the dictionary
        value
            The value to return if the key is not found. If not specified, None is returned.

        Returns
        -------
            The value of the key.

        """
        try:
            return self[key]
        except:
            return value

    def values(self):
        """It takes the dictionary, enumerates the keys, and then creates a new dictionary with the
        enumerated keys as the keys and the values as the values. Then it returns the values of the new
        dictionary

        Returns
        -------
            The values of the dictionary.

        """
        values = {i: self[key] for i, key in enumerate(self.keys())}
        return values.values()

    def items(self):
        """It returns a list of tuples, where each tuple is a key-value pair

        Returns
        -------
            A dictionary of the keys and values of the class.

        """
        items = {
            key: value for key, value in zip(list(self.keys()), list(self.values()))
        }
        return items.items()

    def update(self, other: Dict[str, Any]):
        """If the value is a dictionary, then create a new Document object with the value as the data.
        Otherwise, just set the value

        Parameters
        ----------
        other : Dict[str, Any]
            Dict[str, Any]

        """
        for key, value in other.items():
            if isinstance(value, dict):
                self.data[key] = Document(value)
            else:
                self.data[key] = value


class DocumentList(DocUtils, MutableSequence):
    """
    A Class for handling json like arrays of dictionaries

    Example:
    >>> docs = DocumentList([{"value": 2}, {"value": 10}])
    """

    def __init__(self, documents: List):
        """If the document is not an instance of Document, then create a new Document object with the
        document as the argument. Otherwise, just use the document as is

        Parameters
        ----------
        documents : List
            List

        """
        super().__init__()

        self.documents = [
            Document(document) if not isinstance(document, Document) else document
            for document in documents
        ]

    def _ipython_display_(self):
        """> The function takes a JSON object and displays it in a pretty format"""
        display(self.json())

    def __repr__(self):
        """It takes a list of documents and returns a list of dictionaries, where each dictionary
        represents a document

        Returns
        -------
            A list of dictionaries.

        """
        return repr([document.json() for document in self.documents])

    def __len__(self):
        """The function returns the length of the documents list

        Returns
        -------
            The length of the documents list.

        """
        return len(self.documents)

    def __add__(self, other):
        """The function takes two objects of the same class and adds the documents of the second object to
        the first object

        Parameters
        ----------
        other
            The other object to add to this one.

        Returns
        -------
            The documents are being returned.

        """
        self.documents += other.documents
        return self

    def __contains__(self, document):
        """The function returns True if the document is in the corpus, and False otherwise

        Parameters
        ----------
        document
            The document to check for.

        Returns
        -------
            The document is being returned.

        """
        return document in self.documents

    def __getitem__(self, index):
        """This function returns the document at the given index

        Parameters
        ----------
        index
            The index of the document to be retrieved.

        Returns
        -------
            The documents in the corpus.

        """
        return self.documents[index]

    def __setitem__(self, index, document):
        """It takes a document and adds it to the documents list

        Parameters
        ----------
        index
            The index of the document to be updated.
        document
            The document to be inserted.

        """
        if isinstance(document, dict):
            document = Document(document)
        assert isinstance(document, Document)
        self.documents[index] = document

    def __delitem__(self, index):
        """This function deletes the document at the given index

        Parameters
        ----------
        index
            The index of the item to be removed.

        """
        del self.documents[index]

    def insert(self, index, document):
        """It takes a document and inserts it into the list of documents at the specified index

        Parameters
        ----------
        index
            The index of the element before which to insert the new element.
        document
            The document to insert.

        """
        if isinstance(document, dict):
            document = Document(document)
        assert isinstance(document, Document)
        self.documents.insert(index, document)

    def json(self):
        """It returns a list of json objects.

        Returns
        -------
            A list of dictionaries.

        """
        return [document.json() for document in self.documents]



from relevanceai.operations_new.processing.text.translate.transform import (
    TranslateTransform,
)
from relevanceai.operations_new.ops_base import OperationAPIBase


class TranslateOps(TranslateTransform, OperationAPIBase):
    def __init__(
        self, credentials, fields: list, model_id: str = None, *args, **kwargs
    ):
        self.fields = fields
        self.model_id = model_id
        self.credentials = credentials
        super().__init__(fields, model_id)

    @property
    def name(self):
        return "translate"



import traceback
from typing import List, Dict, Any, Optional

from relevanceai.dataset import Dataset
from relevanceai.operations_new.ops_run import OperationRun

from tqdm.auto import tqdm


class OperationManager:
    """
    Operation manager manages an operation.
    It handles storing operation manager.

    In future, it will also handle:
    - logging
    - timing
    """

    def __init__(
        self,
        dataset: Dataset,
        operation: OperationRun,
        metadata: Optional[Dict[str, Any]] = None,
        post_hooks: Optional[list] = None,
    ):
        self.dataset = dataset
        self.operation = operation
        self.metadata = metadata
        self.post_hooks = post_hooks if post_hooks is not None else []

    def __enter__(self):
        return self.dataset

    def __exit__(self, *args, **kwargs):
        from relevanceai.operations_new.cluster.ops import ClusterOps
        from relevanceai.operations_new.cluster.sub.ops import SubClusterOps
        from relevanceai.operations_new.cluster.batch.ops import BatchClusterOps

        if isinstance(self.operation, SubClusterOps):
            # Get the centroids if there are any
            # Once they are retrieved, we want to get the centroids of the new subcluster operation
            # Then we want to update the original centroids
            # Then upsert them back into the centroids dictionary
            try:
                from relevanceai.utils import suppress_stdout_stderr

                with suppress_stdout_stderr():
                    centroids = self.operation.centroids
                centroid_documents = {
                    d["_id"]: d[self.operation.vector_field] for d in centroids
                }
            except:
                centroid_documents = {}
            subcluster_centroid_documents = self.operation.get_centroid_documents()
            subcluster_centroid_documents = {
                d["_id"]: d[self.operation.vector_fields[0]]
                for d in subcluster_centroid_documents
            }
            centroid_documents.update(subcluster_centroid_documents)
            centroid_documents = [
                {"_id": k, self.operation.vector_fields[0]: v}
                for k, v in centroid_documents.items()
            ]

        elif isinstance(self.operation, ClusterOps):
            centroid_documents = self.operation.get_centroid_documents()

        elif isinstance(self.operation, BatchClusterOps):
            centroid_documents = self.operation.get_centroid_documents()

        if "centroid_documents" in locals():
            tqdm.write("Inserting centroids...")
            res = self.operation.insert_centroids(centroid_documents)
            tqdm.write("Centroids Intserted!")

        for h in self.post_hooks:
            h()

        self._store_default_parent_child()

        self.operation.store_operation_metadata(
            dataset=self.dataset,
            values=self.metadata,
        )

    def _store_default_parent_child(self):
        """We temporarily store the default parent child relationship
        where possible
        """
        try:
            if hasattr(self.operation, "output_fields"):
                if hasattr(self.operation, "select_fields"):
                    for i, field in enumerate(self.operation.select_fields):
                        self.dataset.update_field_children(
                            field=field,
                            field_children=[self.operation.output_fields[i]],
                            category=self.operation.name,  # Should this be the workflow ID
                            metadata={},
                        )
                elif hasattr(self.operation, "fields"):
                    for i, field in enumerate(self.operation.fields):
                        self.dataset.update_field_children(
                            field=field,
                            field_children=[self.operation.output_fields[i]],
                            category=self.operation.name,  # Should this be the workflow ID
                            metadata={},
                        )
                elif hasattr(self.operation, "text_fields"):
                    for i, field in enumerate(self.operation.text_fields):
                        self.dataset.update_field_children(
                            field=field,
                            field_children=[self.operation.output_fields[i]],
                            category=self.operation.name,  # Should this be the workflow ID
                            metadata={},
                        )
        except Exception as e:
            # TODO: rigorously test this with different operations
            # reason: `output_fields` are all calculated differently in
            # different spots for different operations

            traceback.print_exc()

    @staticmethod
    def clean(
        before_docs: List[Dict[str, Any]],
        after_docs: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        updated_documents = [
            {
                key: value
                for key, value in after_doc.items()
                if key not in before_doc or key == "_id"
            }
            for (before_doc, after_doc,) in zip(
                before_docs,
                after_docs,
            )
        ]
        return updated_documents



import numpy as np
from relevanceai.operations.dr.base import DimReductionBase
from typing import Optional, Dict, Any
from relevanceai.operations.cluster.constants import (
    DIM_REDUCTION,
    DIM_REDUCTION_DEFAULT_ARGS,
)


class UMAP(DimReductionBase):
    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["umap"],
        dims: int = 3,
    ) -> np.ndarray:
        try:
            from umap import UMAP
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall umap\n \
                pip install -U relevanceai[umap]"
            )
        self.logger.debug(f"{dr_args}")
        umap = UMAP(n_components=dims, **dr_args)
        return umap.fit_transform(vectors)



from copy import deepcopy

from relevanceai.operations_new.transform_base import TransformBase

from relevanceai.operations_new.processing.text.clean.helpers import (
    MLStripper,
    BaseTextProcessing,
)


class CleanTextTransform(TransformBase, BaseTextProcessing):
    def __init__(self):
        self.stripper = MLStripper()

    @property
    def name(self):
        return "clean_text"

    def clean_text(self, text):
        """
        Clean the text of the individuals
        """
        text = CleanTextTransform.normalize_text(
            text, lower=False, remove_punct=False, remove_digit=False
        )
        return self.stripper.clean(text)

    def clean_text_document(
        self, text_field, document, output_field: str = "_cleantext_"
    ):
        """
        Split a text field and store it in other values
        """
        t = self.get_field(text_field, document)
        clean_text = self.clean_text(t)
        # Format the split text into documents
        self.set_field(output_field + "." + text_field, document, clean_text)
        return document

    def transform(
        self,
        text_field,
        documents,
        inplace: bool = True,
        output_field: str = "_cleantext_",
    ):
        if not inplace:
            documents = deepcopy(documents)
        # TODO; switch to something faster than list comprehension
        [
            self.clean_text_document(
                text_field=text_field, document=document, output_field=output_field
            )
            for document in documents
        ]
        return documents



import warnings
from typing import Any, Optional
from relevanceai.constants.warning import Warning


class ClusterAlias:
    """Utilities for determining a cluster alias"""

    model: Any
    model_kwargs: dict
    model_name: str
    verbose: bool = False

    def _get_alias(self, alias: Any) -> str:
        # Return the alias
        if alias is not None:
            self.alias = alias
            return alias
        alias = self._get_alias_from_package()

        if alias is not None and isinstance(alias, str):
            return alias
        alias = self._generate_alias()
        return alias.lower()

    def _get_alias_from_package(self):
        self._get_package_from_model(self.model)
        if self.package == "sklearn":
            self.alias = self._get_alias_from_sklearn()
            if self.alias is not None:
                return self.alias

    def _get_alias_from_sklearn(self):
        if hasattr(self.model, "name"):
            if hasattr(self.model, "n_clusters"):
                return f"{self.model.name}-{self.model.n_clusters}"
            elif hasattr(self.model, "k"):
                return f"{self.model.name}-{self.model.k}"
            else:
                return f"{self.model.name}"
        else:
            warnings.warn("No alias has been detected - using model type.")
            return str(type(self.model))

    def _get_n_clusters(self):
        if "n_clusters" in self.model_kwargs:
            return self.model_kwargs["n_clusters"]
        elif hasattr(self.model, "n_clusters"):
            return self.model.n_clusters
        elif hasattr(self.model, "k"):
            return self.model.k
        return None

    def _generate_alias(self) -> str:
        # Issue a warning about auto-generated alias
        # We auto-generate certain aliases if the model
        # is a default model like kmeans or community detection
        n_clusters = self._get_n_clusters()
        if hasattr(self.model, "alias"):
            alias = self.model.alias
            if alias is not None:
                return alias

        if n_clusters is not None:
            alias = f"{self.model_name}-{n_clusters}"
        else:
            alias = f"{self.model_name}"

        if self.verbose:
            print(f"The alias is `{alias.lower()}`.")

        Warning.MISSING_ALIAS.format(alias=alias)
        return alias

    def _get_package_from_model(self, model):
        """
        Determine the package for a model.
        This can be useful for checking dependencies.
        This may be used across modules for
        deeper integrations
        """
        # TODO: add support for huggingface integrations
        # such as transformers and sentencetransformers
        model_name = str(model.__class__).lower()
        if "function" in model_name:
            model_name = str(model.__name__)

        if "sklearn" in model_name:
            self.package = "sklearn"

        elif "faiss" in model_name:
            self.package = "faiss"

        elif "hdbscan" in model_name:
            self.package = "hdbscan"

        elif "communitydetection" in model_name:
            self.package = "sentence-transformers"

        else:
            self.package = "custom"
        return self.package

    def _get_model(self, model: Any, model_kwargs: Optional[dict]) -> Any:
        if model_kwargs is None:
            model_kwargs = {}

        if model is None:
            return model
        if isinstance(model, str):
            model = self._get_model_from_string(model, model_kwargs)

        elif "sklearn" in model.__module__:
            model = self._get_sklearn_model_from_class(model)

        elif "faiss" in model.__module__:
            model = self._get_faiss_model_from_class(model)

        return model

    def _get_sklearn_model_from_class(self, model):
        from relevanceai.operations_new.cluster.models.sklearn.base import (
            SklearnModel,
        )

        model_kwargs = model.__dict__
        model = SklearnModel(model=model, model_kwargs=model_kwargs)
        return model

    def _get_faiss_model_from_class(self, model):
        raise NotImplementedError

    def normalize_model_name(self, model):
        if isinstance(model, str):
            return model.lower().replace("-", "").replace("_", "")

        return model

    def _get_model_from_string(self, model: str, model_kwargs: dict = None):
        if model_kwargs is None:
            model_kwargs = {}

        model = self.normalize_model_name(model)
        model_kwargs = {} if model_kwargs is None else model_kwargs

        from relevanceai.operations_new.cluster.models.sklearn import sklearn_models

        if model in sklearn_models:
            from relevanceai.operations_new.cluster.models.sklearn.base import (
                SklearnModel,
            )

            model = SklearnModel(
                model=model,
                model_kwargs=model_kwargs,
            )
            return model

        elif model == "faiss":
            from relevanceai.operations_new.cluster.models.faiss.base import (
                FaissModel,
            )

            model = FaissModel(
                model=model,
                model_kwargs=model_kwargs,
            )
            return model

        elif model == "communitydetection":
            from relevanceai.operations_new.cluster.models.sentence_transformers.community_detection import (
                CommunityDetectionModel,
            )

            model = CommunityDetectionModel(**model_kwargs)
            return model

        raise ValueError("Model not supported.")



import numpy as np
from relevanceai.operations.dr.base import DimReductionBase
from typing import Optional, Dict, Any
from relevanceai.operations.cluster.constants import (
    DIM_REDUCTION,
    DIM_REDUCTION_DEFAULT_ARGS,
)


class Ivis(DimReductionBase):
    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["ivis"],
        dims: int = 3,
    ) -> np.ndarray:
        try:
            from ivis import Ivis
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall ivis\n \
                CPU: pip install -U ivis-cpu\n \
                GPU: pip install -U ivis-gpu"
            )
        self.logger.debug(f"{dr_args}")
        ivis = Ivis(embedding_dims=dims, **dr_args)
        if ivis.chunksize > vectors.shape[0]:
            ivis.chunksize = vectors.shape[0]
        vectors_dr = ivis.fit(vectors).transform(vectors)
        return vectors_dr



import itertools
from typing import Dict, List, Optional, Union
from relevanceai.apps.report_app.advanced_blocks import ReportAdvancedBlocks


class AggregateSections(ReportAdvancedBlocks):
    def section_overview(
        self,
        metrics: List,
        entity_name: str = "documents",
        filters: List = None,
        decimals: int = 3,
    ):
        filters = [] if filters is None else filters

        main_metrics, metric_fields, _ = self.dataset._clean_metrics(metrics)
        self.quote(
            f"Analyzing {self.dataset.shape[0]} {entity_name} across {self.dataset.shape[1]} fields/columns"
        )
        self.h1("Section: Overview")
        overall_metrics = self.dataset.aggregate(metrics=main_metrics, filters=filters)[
            "results"
        ][0]
        overall_metrics_bullets = []
        for k, v in overall_metrics.items():
            if k == "frequency":
                continue
            overall_metrics_bullets += [
                [self.bold(k), ": ", self.bold(str(round(v, decimals)))]
            ]
        self.bullet_list(overall_metrics_bullets)

    def section_groupby(
        self,
        groupby: List,
        metrics: List = None,
        groupby_depths: List = None,
        filters: List = None,
        decimals: int = 3,
    ):
        metrics = [] if metrics is None else metrics
        groupby_depths = [1] if groupby_depths is None else groupby_depths
        filters = [] if filters is None else filters

        self.h1("Section: Analyzing different groups")
        main_metrics, metric_fields, _ = self.dataset._clean_metrics(metrics)
        main_groupby, groupby_fields = self.dataset._clean_groupby(groupby)
        for depth in groupby_depths:
            if depth == 1:
                for group in main_groupby:
                    self.h2(f"Looking at ({group['name']})")
                    for order in ["desc", "asc"]:
                        group_freq = self.dataset.aggregate(
                            groupby=[group],
                            metrics=main_metrics,
                            sort=[{"frequency": order}],
                            filters=filters,
                        )["results"]
                        prefix = "Least" if order == "asc" else "Most"
                        freq_str = f"{prefix} frequent {group['name']}: {group_freq[0]['frequency']} for {group_freq[0][group['name']]}"
                        self.paragraph(freq_str)
                    for metric in main_metrics:
                        for order in ["asc", "desc"]:
                            group_metric = self.dataset.aggregate(
                                groupby=[group],
                                metrics=main_metrics,
                                sort=[{metric["name"]: order}],
                                filters=filters,
                            )["results"]
                            prefix = "Lowest" if order == "asc" else "Highest"
                            metric_str = f"{prefix} {metric['name']} for {group['name']}: {round(group_metric[0][metric['name']], decimals)} for {group_metric[0][group['name']]}"
                            self.paragraph(metric_str)
            elif depth > 1:
                for group in list(itertools.combinations(main_groupby, depth)):
                    group_name_str = " & ".join([g["name"] for g in group])
                    self.h2(f"Looking at ({group_name_str})")

                    for order in ["desc", "asc"]:
                        group_freq = self.dataset.aggregate(
                            groupby=list(group),
                            metrics=main_metrics,
                            sort=[{"frequency": order}],
                            filters=filters,
                        )["results"]
                        group_freq_name = [group_freq[0][g["name"]] for g in group]
                        group_freq_name_str = " & ".join(group_freq_name)
                        prefix = "Least" if order == "asc" else "Most"
                        group_freq_str = f"{prefix} frequent for {group_name_str}: {group_freq[0]['frequency']} for {group_freq_name_str}"
                        self.paragraph(group_freq_str)
                    for m in main_metrics:
                        for order in ["desc", "asc"]:
                            group_metric = self.dataset.aggregate(
                                groupby=list(group),
                                metrics=main_metrics,
                                sort=[{m["name"]: order}],
                                filters=filters,
                            )["results"]
                            group_metric_name = [group_metric[0][g["name"]] for g in group]
                            group_metric_name_str = " & ".join(group_metric_name)
                            prefix = "Lowest" if order == "asc" else "Highest"
                            group_str = f"{prefix} {m['name']} for {group_name_str}: {round(group_metric[0][m['name']], decimals)} for {group_metric_name_str}"
                            self.paragraph(group_str)


#



import sys
import uuid


def make_id(document):
    _id = str(uuid.uuid3(uuid.NAMESPACE_DNS, str(document)))
    return _id


def str2bool(v: str):
    return v.lower() in ("yes", "true", "t", "1")


def getsizeof(obj, seen=None):
    """Recursively finds size of objects"""

    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()

    obj_id = id(obj)
    if obj_id in seen:
        return 0

    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)

    if isinstance(obj, dict):
        size += sum([getsizeof(v, seen) for v in obj.values()])
        size += sum([getsizeof(k, seen) for k in obj.keys()])
    elif hasattr(obj, "__dict__"):
        size += getsizeof(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([getsizeof(i, seen) for i in obj])

    return size



from sklearn.cluster import KMeans
from relevanceai.apps.recipes.model_observability.cluster.evaluation import ClusterEvaluation

def elbow_method(
    X, 
    clusters_to_try=range(1, 10), 
    full_evaluation:bool=False, 
    **kwargs
):
    for k in clusters_to_try:
        model = KMeans(n_clusters=k, **kwargs)
        cluster_labels = model.fit(X)
        ClusterEvaluation(X, cluster_labels)
    return 

def dendrogram_method(X):
    return 


from typing import Dict, List, Any

from relevanceai.operations_new.vectorize.transform import VectorizeTransform
from relevanceai.operations_new.vectorize.models.base import VectorizeModelBase
from relevanceai.operations_new.vectorize.models.image.mappings import *


class VectorizeImageTransform(VectorizeTransform):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_model(self, model: Any) -> VectorizeModelBase:
        """If the model is a string, then check if it's in the TFHUB_MODELS dictionary. If it is, then
        return a TFHubImage2Vec object. If it's not, then raise a ValueError

        Parameters
        ----------
        model : Any
            Any

        Returns
        -------
            The model is being returned.

        """

        if isinstance(model, str):
            if model in TFHUB_MODELS:
                from relevanceai.operations_new.vectorize.models.image.tfhub import (
                    TFHubImage2Vec,
                )

                vector_length = TFHUB_MODELS[model]["vector_length"]
                url = TFHUB_MODELS[model]["url"]

                image_dimensions = (
                    TFHUB_MODELS[model]["image_dimensions"]
                    if "image_dimensions" in TFHUB_MODELS[model]
                    else None
                )
                model = TFHubImage2Vec(
                    url=url,
                    vector_length=vector_length,
                    image_dimensions=image_dimensions,
                )

                return model

            if model in CLIP_MODELS:
                from relevanceai.operations_new.vectorize.models.image.clip import (
                    ClipImage2Vec,
                )

                vector_length = CLIP_MODELS[model]["vector_length"]
                url = CLIP_MODELS[model]["url"]

                model = ClipImage2Vec(
                    url=url,
                    vector_length=vector_length,
                )

                return model

            else:
                raise ValueError("Model not a valid model string")

        elif isinstance(model, VectorizeModelBase):
            return model

        else:
            raise ValueError(
                "Model should either be a supported model string or inherit from ModelBase"
            )



from relevanceai.utils.decorators.analytics import track
from relevanceai.constants import EXPLORER_APP_LINK
from relevanceai.dataset.apps.manage_apps import ManageApps


class LaunchApps(ManageApps):
    """
    Launch apps are designed to create apps automatically.
    """

    @track
    def launch_explore_app(self):
        print(EXPLORER_APP_LINK.format(self.dataset_id))

    @track
    def launch_search_app(self):
        print(EXPLORER_APP_LINK.format(self.dataset_id))



"""
All operations related to sentiment
"""
from typing import List
from relevanceai.operations_new.sentiment.transform import SentimentTransform
from relevanceai.operations_new.ops_base import OperationAPIBase


class SentimentOps(SentimentTransform, OperationAPIBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_filters(text_fields: List[str], model_name: str) -> List[str]:

        output_fields = [
            f"_sentiment_.{text_field}.{model_name}.sentiment"
            for text_field in text_fields
        ]

        if len(text_fields) > 1:
            iters = len(text_fields) ** 2

            filters: list = []
            for i in range(iters):
                binary_array = [character for character in str(bin(i))][2:]
                mixed_mask = ["0"] * (
                    len(text_fields) - len(binary_array)
                ) + binary_array
                mask = [int(value) for value in mixed_mask]
                # Creates a binary mask the length of fields provided
                # for two fields, we need 4 iters, going over [(0, 0), (1, 0), (0, 1), (1, 1)]

                condition_value = [
                    {
                        "field": field if mask[index] else output_field,
                        "filter_type": "exists",
                        "condition": "==" if mask[index] else "!=",
                        "condition_value": "",
                    }
                    for index, (field, output_field) in enumerate(
                        zip(text_fields, output_fields)
                    )
                ]
                filters += [{"filter_type": "or", "condition_value": condition_value}]

        else:  # Special Case when only 1 field is provided
            condition_value = [
                {
                    "field": text_fields[0],
                    "filter_type": "exists",
                    "condition": "==",
                    "condition_value": " ",
                },
                {
                    "field": f"_sentiment_.{text_fields[0]}.{model_name}.sentiment",
                    "filter_type": "exists",
                    "condition": "!=",
                    "condition_value": " ",
                },
            ]
            filters = condition_value

        return filters



import numpy as np

from relevanceai.constants import MissingPackageError


class CommunityDetection:
    def __init__(
        self,
        threshold=0.75,
        min_community_size=10,
        init_max_size=1000,
        gpu=False,
    ):

        self.gpu = gpu
        self.threshold = threshold
        self.min_community_size = min_community_size
        self.init_max_size = init_max_size

    def __call__(self, *args, **kwargs):
        return self.fit_predict(*args, **kwargs)

    def fit_predict(self, vectors):
        if self.gpu:
            communities = self.community_detection_gpu(vectors)
        else:
            communities = self.community_detection_cpu(vectors)

        labels = [-1 for _ in range(vectors.shape[0])]
        for cluster_index, community in enumerate(communities):
            for index in community:
                labels[index] = cluster_index

        return np.array(labels)

    def cosine(self, embeddings):
        """
        effecient cosine sim
        """
        similarity = np.dot(embeddings, embeddings.T)
        square_mag = np.diag(similarity)
        inv_square_mag = 1 / square_mag
        inv_square_mag[np.isinf(inv_square_mag)] = 0
        inv_mag = np.sqrt(inv_square_mag)
        cosine = similarity * inv_mag
        cosine = cosine.T * inv_mag
        return cosine

    def topk(self, embeddings, k):
        """
        numpy topk
        """
        if len(embeddings.shape) == 1:
            embeddings = embeddings.reshape(1, -1)
        indices = embeddings.argpartition(-k, axis=1)[:, -k:]
        indices = np.flip(indices, 1)

        values = np.flip(
            np.sort(embeddings[np.indices(indices.shape)[0], indices], 1), 1
        )
        # TODO: optimise this somehow
        indices = np.array(
            [
                [embeddings[col_index].tolist().index(v) for v in row]
                for col_index, row in enumerate(values.tolist())
            ]
        )
        return values, indices

    def community_detection_cpu(self, embeddings):
        self.init_max_size = min(self.init_max_size, len(embeddings))
        cos_scores = self.cosine(embeddings)
        top_k_values, _ = self.topk(cos_scores, k=self.min_community_size)

        extracted_communities = []
        for i in range(len(top_k_values)):
            if top_k_values[i][-1] >= self.threshold:
                new_cluster = []

                top_val_large, top_idx_large = self.topk(
                    cos_scores[i], k=self.init_max_size
                )
                top_idx_large = top_idx_large.flatten().tolist()
                top_val_large = top_val_large.flatten().tolist()

                if top_val_large[-1] < self.threshold:
                    for idx, val in zip(top_idx_large, top_val_large):
                        if val < self.threshold:
                            break

                        new_cluster.append(idx)
                else:
                    for idx, val in enumerate(cos_scores[i].tolist()):
                        if val >= self.threshold:
                            new_cluster.append(idx)

                extracted_communities.append(new_cluster)

        extracted_communities = sorted(
            extracted_communities, key=lambda x: len(x), reverse=True
        )

        unique_communities = []
        extracted_ids = set()

        for community in extracted_communities:
            add_cluster = True
            for idx in community:
                if idx in extracted_ids:
                    add_cluster = False
                    break

            if add_cluster:
                unique_communities.append(community)
                for idx in community:
                    extracted_ids.add(idx)

        print(f"There were {len(unique_communities)} communities found.")
        return unique_communities

    def community_detection_gpu(self, embeddings):
        try:
            from sentence_transformers.util import community_detection
        except ModuleNotFoundError:
            raise MissingPackageError("sentence-transformers")
        return community_detection(
            embeddings,
            self.threshold,
            self.min_community_size,
            self.init_max_size,
        )



"""
Labelling with API-related functions
"""
from re import I
from relevanceai.operations_new.text_tagging.transform import TextTagTransform
from relevanceai.operations_new.ops_base import OperationAPIBase


class TextTagOps(TextTagTransform, OperationAPIBase):  # type: ignore
    """
    Label Operations
    """

    def __init__(
        self, credentials, fields, labels, output_fields, model_id=None, **kwargs
    ):
        self.credentials = credentials
        self.fields = fields
        self.text_field = fields[0]
        self.labels = labels
        if len(fields) > 1:
            raise ValueError("cannot support more than 1 field.")
        if model_id is None:
            self.model_id = "cross-encoder/nli-deberta-v3-large"
        else:
            self.model_id = model_id
        if output_fields is None:
            self.output_fields = [self._generate_output_field(f) for f in fields]
        else:
            self.output_fields = output_fields
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.model_id = model_id



import requests

from typing import List
from concurrent.futures import ThreadPoolExecutor

from relevanceai.operations_new.vectorize.models.base import VectorizeModelBase
from relevanceai.utils.decorators.vectors import catch_errors

try:
    import traceback
    import clip
    import torch
    import requests
    import cv2
    from PIL import Image
    from requests.exceptions import MissingSchema
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "Run `pip install git+https://github.com/openai/CLIP.git`."
    )
except:
    import traceback

    traceback.print_exc()


class ClipImage2Vec(VectorizeModelBase):
    def __init__(self, url, vector_length, context_length=77):

        self.context_length = context_length
        self.device = "cuda" if torch.cuda.is_available else "cpu"
        self.model, self.preprocess = clip.load(url, device=self.device)
        self.vector_length = vector_length
        self.url = url

    @property
    def model_name(self):
        return "clip"

    def read(self, image_url):
        """It takes an image url, and returns an image object

        Parameters
        ----------
        image_url
            The URL of the image to be downloaded.

        Returns
        -------
            The image is being returned.

        """
        try:
            return Image.open(requests.get(image_url, stream=True).raw)
        except MissingSchema:
            return Image.open(image_url)

    def preprocess_black_and_white_image(self, x):
        """It takes a black and white image, converts it to a tensor, adds two more channels to it, and
        then normalizes it

        Parameters
        ----------
        x
            the image

        Returns
        -------
            A tensor of shape (3, 224, 224)

        """
        x = self.preprocess.transforms[0](x)
        x = self.preprocess.transforms[1](x)
        x = self.preprocess.transforms[3](x)
        x = torch.stack((x, x, x), dim=1)
        x = self.preprocess.transforms[4](x)
        return x

    @catch_errors
    def encode_text(self, text: str):
        """If the device is cuda, then tokenize the text, send it to the device, encode it, detach it,
        convert it to numpy, convert it to a list, and return the first element of that list

        Parameters
        ----------
        text : str
            str

        Returns
        -------
            A list of floats.

        """
        if self.device == "cuda":
            text = clip.tokenize(text, context_length=self.context_length).to(
                self.device
            )
            return self.model.encode_text(text).cpu().detach().numpy().tolist()[0]
        elif self.device == "cpu":
            text = clip.tokenize(text, context_length=self.context_length).to(
                self.device
            )
            return self.model.encode_text(text).detach().numpy().tolist()[0]

    def encode_video(self, video_url: str):
        """It takes a video file, reads it frame by frame, converts each frame to a PIL image, preprocesses
        it, converts it to a tensor, and then encodes it using the model

        Parameters
        ----------
        video_url : str
            the path to the video file

        Returns
        -------
            A list of floats

        """
        cap = cv2.VideoCapture(video_url)
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            pil_img = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        image = self.preprocess(pil_img).unsqueeze(0).to(self.device)
        return self.model.encode_image(image).cpu().detach().numpy().tolist()[0]

    def bulk_encode_text(self, texts: List[str]):
        """If the device is cuda, then tokenize the text, send it to the device, encode it, and return the
        result.

        If the device is cpu, then tokenize the text, send it to the device, encode it, and return the
        result.

        The only difference is the device.

        So, let's make a function that takes a device as an argument and returns a function that does
        the above.

        We'll call this function `make_encode_text_function`.

        The function it returns will be called `encode_text`.

        We'll use the `partial` function from the `functools` module to do this.

        The `partial` function takes a function and some arguments and returns a function that takes the
        remaining arguments.

        So, we'll pass it the `bulk_encode_text` function and the `device

        Parameters
        ----------
        texts : List[str]
            List[str]

        Returns
        -------
            A list of lists of floats.

        """
        if self.device == "cuda":
            tokenized_text = clip.tokenize(
                texts, context_length=self.context_length
            ).to(self.device)
            return (
                self.model.encode_text(tokenized_text).cpu().detach().numpy().tolist()
            )
        elif self.device == "cpu":
            tokenized_text = clip.tokenize(
                texts, context_length=self.context_length
            ).to(self.device)
            return self.model.encode_text(tokenized_text).detach().numpy().tolist()

    def preprocess_image(self, img: str):
        try:
            if self.is_greyscale(img):
                return (
                    self.preprocess_black_and_white_image(self.read(img))
                    .unsqueeze(0)
                    .to(self.device)
                )
            return self.preprocess(self.read(img)).unsqueeze(0).to(self.device)
        except:
            traceback.print_exc()
            return torch.empty((1, 3, 224, 224), dtype=torch.int32, device=self.device)

    def parallel_preprocess_image(self, images: List[str]):
        with ThreadPoolExecutor(max_workers=5) as executor:
            future = executor.map(self.preprocess_image, images)
        return list(future)

    @catch_errors
    def encode_image(self, image_url: str):
        """If the device is cpu, then preprocess the image, add a dimension to the image, and then return
        the encoded image.

        If the device is cuda, then preprocess the image, add a dimension to the image if it's 3D, and
        then return the encoded image

        Parameters
        ----------
        image_url : str
            The URL of the image to be encoded.

        Returns
        -------
            A list of floats

        """
        if self.device == "cpu":
            image = self.preprocess_image(image_url)
            if image.dim() == 3:
                image = image.unsqueeze(0).to(self.device)
            return self.model.encode_image(image).detach().numpy().tolist()[0]
        elif self.device == "cuda":
            image = self.preprocess_image(image_url)
            if image.ndim == 3:
                image = image.unsqueeze(0).to(self.device)
            elif image.ndim == 4:
                image = image.to(self.device)
            return self.model.encode_image(image).cpu().detach().numpy().tolist()[0]

    def bulk_encode_image(self, images: List[str]):
        """Batch Processing for CLIP image encoding"""
        # Parallel process the encoding
        future = self.parallel_preprocess_image(images)
        results = self.model.encode_image(torch.cat(list(future))).tolist()
        return results

    def encode(self, data: str, data_type="image"):
        if data_type == "image":
            return self.encode_image(data)
        elif data_type == "text":
            return self.encode_text([data])
        raise ValueError("data_type must be either `image` or `text`")

    def bulk_encode(self, data: List[str], data_type="image"):
        if data_type == "image":
            return self.bulk_encode_image(data)
        elif data_type == "text":
            return self.bulk_encode_text(data)
        raise ValueError("data_type must be either `image` or `text`")

    def is_greyscale(self, img_path: str):
        """Determine if an image is grayscale or not"""
        try:
            img = Image.open(requests.get(img_path, stream=True).raw)
        except MissingSchema:
            img = Image.open(img_path)
        img = img.convert("RGB")
        w, h = img.size
        for i in range(w):
            for j in range(h):
                r, g, b = img.getpixel((i, j))
                if r != g != b:
                    return False
        return True



from typing import Optional

from relevanceai.dataset.write.write import Write
from relevanceai.utils.decorators.analytics import track


# TODO: Separate out operations into different files - cluster/search/dr
class DimensionalityReduction(Write):
    def _run_dr_algorithm(
        self,
        algorithm: str,
        vector_fields: list,
        documents: list,
        alias: str,
        n_components: int = 3,
    ):
        original_name = algorithm
        # Make sure that the letter case does not matter
        algorithm = algorithm.upper()
        if algorithm == "PCA":
            from relevanceai.operations.dr.models import PCA

            model = PCA()
        elif algorithm == "TSNE":
            from relevanceai.operations.dr.models import TSNE

            model = TSNE()
        elif algorithm == "UMAP":
            from relevanceai.operations.dr.models import UMAP

            model = UMAP()
        elif algorithm == "IVIS":
            from relevanceai.operations.dr.models import Ivis

            model = Ivis()
        else:
            raise ValueError(
                f'"{original_name}" is not a supported '
                "dimensionality reduction algorithm. "
                "Currently, the supported algorithms are: "
                "PCA, TSNE, UMAP, and IVIS"
            )

        print(f"Run {algorithm}...")
        # Returns a list of documents with dr vector
        return model.fit_transform_update(
            vector_field=vector_fields[0],
            documents=documents,
            alias=alias,
            dims=n_components,
        )

    @track
    def auto_reduce_dimensions(
        self,
        alias: str,
        vector_fields: list,
        filters: Optional[list] = None,
        number_of_documents: Optional[int] = None,
    ):
        """
        Run dimensionality reduction quickly on a dataset on a small number of documents.
        This is useful if you want to quickly see a projection of your dataset.

        .. warning::
            This function is currently in beta and is likely to change in the future.
            We recommend not using this in any production systems.

        Parameters
        ----------
        vector_fields: list
            The vector fields to run dimensionality reduction on
        number_of_documents: int
            The number of documents to get
        algorithm: str
            The algorithm to run. The only supported algorithm is `pca` at this
            current point in time.
        n_components: int
            The number of components

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            df.auto_reduce_dimensions(
                "pca-3",
                ["sample_vector_"],
            )

        """
        if len(vector_fields) > 1:
            raise ValueError("We only support 1 vector field at the moment.")

        dr_args = alias.split("-")

        if len(dr_args) != 2:
            raise ValueError("""Your DR alias should be in the form of `pca-3`.""")

        algorithm = dr_args[0]
        n_components = int(dr_args[1])

        print("Getting documents...")
        if filters is None:
            filters = []
        filters += [
            {
                "field": vf,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for vf in vector_fields
        ]

        if number_of_documents is None:
            number_of_documents = self.get_number_of_documents(self.dataset_id, filters)

        documents = self.get_documents(
            select_fields=vector_fields,
            filters=filters,
            number_of_documents=number_of_documents,
        )

        dr_documents = self._run_dr_algorithm(
            algorithm=algorithm,
            vector_fields=vector_fields,
            documents=documents,
            alias=alias,
            n_components=n_components,
        )

        results = self.update_documents(self.dataset_id, dr_documents)

        if n_components == 3:
            projector_url = f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/projector"
            print(f"You can now view your projector at {projector_url}")

        return results

    @track
    def reduce_dimensions(
        self,
        vector_fields: list,
        alias: str,
        number_of_documents: int = 1000,
        algorithm: str = "pca",
        n_components: int = 3,
        filters: Optional[list] = None,
    ):
        """
        Run dimensionality reduction quickly on a dataset on a small number of documents.
        This is useful if you want to quickly see a projection of your dataset.

        .. warning::
            This function is currently in beta and is likely to change in the future.
            We recommend not using this in any production systems.

        Parameters
        ----------
        vector_fields: list
            The vector fields to run dimensionality reduction on
        number_of_documents: int
            The number of documents to get
        algorithm: str
            The algorithm to run. The only supported algorithm is `pca` at this
            current point in time.
        n_components: int
            The number of components

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            df.auto_reduce_dimensions(
                alias="pca-3",
                ["sample_vector_"],
                number_of_documents=1000
            )

        """
        filters = [] if filters is None else filters

        if len(vector_fields) > 1:
            raise ValueError("We only support 1 vector field at the moment.")

        print("Getting documents...")
        filters += [
            {
                "field": vf,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for vf in vector_fields
        ]
        documents = self.get_documents(
            select_fields=vector_fields,
            filters=filters,
            number_of_documents=number_of_documents,
        )

        dr_documents = self._run_dr_algorithm(
            algorithm=algorithm,
            vector_fields=vector_fields,
            documents=documents,
            alias=alias,
            n_components=n_components,
        )

        results = self.update_documents(self.dataset_id, dr_documents)

        return results



"""
Sentence Splitting Operator
Split Text
"""
from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.processing.text.ner.transform import ExtractNER


class ExtractNEROps(OperationAPIBase, ExtractNER):
    def __init__(
        self,
        credentials,
        fields,
        output_fields,
        model_id: str = "dslim/bert-base-NER",
        **kwargs
    ):
        self.credentials = credentials
        self.fields = fields
        if model_id is None:
            self.model_id = "dslim/bert-base-NER"
        if output_fields is None:
            self.output_fields = [self._generate_output_field(f) for f in fields]
        else:
            self.output_fields = output_fields
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.model_id = model_id



"""
Test that analyzing text works out of the box
"""
import pytest
from relevanceai.dataset import Dataset
from typing import List


@pytest.mark.parametrize(
    ["vectorize_models", "subcluster_model", "filters"],
    [
        (
            ["princeton-nlp/sup-simcse-roberta-large"],
            "all-mpnet-base-v2",
            None,
        )
    ],
    ids=["single_model"],
)
def test_analyze_text(
    test_dataset: Dataset,
    vectorize_models: List,
    subcluster_model: str,
    filters: List[dict],
):
    test_vector_name = "sample_vector_"
    test_field = "sample_1_label"

    test_dataset.analyze_text(
        fields=[test_field],
        vector_fields=[test_vector_name],
        vectorize_models=vectorize_models,
        vectorize=True,
        cluster=True,
        subcluster=True,
        subcluster_model=subcluster_model,
        extract_sentiment=True,
        extract_emotion=True,
        count=True,
        filters=filters,
    )
    assert test_vector_name in test_dataset.schema



"""Code for migrating datasets
"""
import math

from typing import Any, Dict, List, Optional

from tqdm.auto import tqdm

from relevanceai.operations_new.ops_run import PullTransformPush


def migrate_dataset(
    old_token: str,
    new_token: str,
    dataset_id: str,
    new_dataset_id: Optional[str] = None,
    chunksize: Optional[int] = None,
    filters: list = None,
    show_progress_bar: bool = True,
    max_workers: int = 1,
):
    """
    Migrate dataset

    Args
    ---------
        old_token (str): _description_
        new_token (str): _description_
        dataset_id (str): _description_
        new_dataset_id (Optional[str], optional): _description_. Defaults to None.
        chunksize (int, optional): _description_. Defaults to 20.

    Example
    ---------

    .. code-block::

        from relevanceai.migration import migrate
        migrate_dataset(
            old_token="...",
            new_token="...",
            dataset_id="sample_dataset",
            new_dataset_id="new_sample_dataset")

    """
    from relevanceai import Client

    filters = filters if filters is not None else []

    if new_dataset_id is None:
        new_dataset_id = dataset_id

    old_client = Client(token=old_token)
    new_client = Client(token=new_token)

    old_dataset = old_client.Dataset(dataset_id)
    new_dataset = new_client.Dataset(dataset_id)

    ptp = PullTransformPush(
        pull_dataset=old_dataset,
        push_dataset=new_dataset,
        func=None,
        show_progress_bar=show_progress_bar,
        show_transform_progress_bar=False,
        pull_chunksize=chunksize,
        push_chunksize=chunksize,
        push_workers=max_workers,
    )
    ptp.run()

    metadata = old_dataset.metadata.to_dict()
    new_dataset.upsert_metadata(metadata)

    cluster_fields = [
        field
        for field in old_dataset.schema
        if "_cluster_" in field and len(field.split(".")) == 3
    ]
    if cluster_fields:
        tqdm.write(f"Found {len(cluster_fields)} cluster_fields")

    for cluster_field in cluster_fields:
        _, vector_field, alias = cluster_field.split(".")

        tqdm.write(f"Inserting Centroids for `{cluster_field}`...")
        old_cluster_ops = old_client.ClusterOps(
            alias=alias,
            vector_fields=[vector_field],
            dataset_id=old_dataset.dataset_id,
        )
        centroid_documents = old_cluster_ops.centroids
        new_dataset.datasets.cluster.centroids.insert(
            dataset_id=new_dataset.dataset_id,
            cluster_centers=new_dataset.json_encoder(centroid_documents),
            vector_fields=[vector_field],
            alias=alias,
        )

    tqdm.write("Centroids inserted!")

    tqdm.write("Migrating Operations Metadata...")
    metadata = old_dataset.metadata.to_dict()
    new_dataset.insert_metadata(metadata)
    tqdm.write("Operations Metadata Migrated!")

    tqdm.write("Finished migrating.")



import requests
from relevanceai.operations_new.transform_base import TransformBase


class TranslateTransform(TransformBase):
    def __init__(
        self,
        fields,
        model_id="facebook/mbart-large-50-many-to-many-mmt",
    ):
        if model_id != "facebook/mbart-large-50-many-to-many-mmt":
            raise NotImplementedError("Translation model not found.")

        self.model_id = model_id
        self.fields = fields
        # Get the language dictionary
        self.lang_dict = requests.get(
            "https://gist.githubusercontent.com/boba-and-beer/b142b44e7c5120c5ce839e4b7f8f0247/raw/21b73269fecb67e871843086fe8ebe78d8b4c7a0/lang_isocode_mapping.json"
        ).json()
        self.lang_to_normal = requests.get(
            "https://gist.githubusercontent.com/boba-and-beer/65bf4a0810d6492050577f1d2373d394/raw/d3817a37401b0776b8ed4bda1bbcf89e8375c747/lang_iso_2_mapping.json"
        ).json()
        from transformers import MBartForConditionalGeneration, MBart50TokenizerFast

        self.model = MBartForConditionalGeneration.from_pretrained(model_id)
        self.tokenizer = MBart50TokenizerFast.from_pretrained(model_id)

    def translate_to_english(self, text):
        from langdetect import detect_langs

        language = detect_langs(text)
        if language[0].lang == "en":
            return None, None
        supported_languages = []
        for lang in language:
            try:
                return self._pass_text_through_model(
                    text, lang.lang, supported_languages
                )
            except Exception as e:
                pass

        return None, None

    def _pass_text_through_model(self, text, language: str, supported_languages: list):
        self.tokenizer.src_lang = self.lang_dict[language]
        if language in self.lang_to_normal:
            supported_languages.append(self.lang_to_normal[language])
        else:
            supported_languages.append(language)

        encoded_hi = self.tokenizer(text, return_tensors="pt")
        generated_tokens = self.model.generate(
            **encoded_hi, forced_bos_token_id=self.tokenizer.lang_code_to_id["en_XX"]
        )
        translation = self.tokenizer.batch_decode(
            generated_tokens, skip_special_tokens=True
        )
        return supported_languages, translation

    def translate_document(self, field, document):
        try:
            sup_lang, translation = self.translate_to_english(
                self.get_field(field, document)
            )
            self.set_field(
                "_translation_." + field,
                document,
                {"detectedLanguage": sup_lang, "translation": translation},
            )
        except:
            pass
        return document

    def bulk_translate_documents(self, documents, **kwargs):
        for field in self.fields:
            [self.translate_document(field, d, **kwargs) for d in documents]
        return documents

    transform = bulk_translate_documents  # type: ignore



import pytest

from relevanceai.utils import DocUtils


def test_is_field(sample_document):
    assert DocUtils().is_field("fehugieh", sample_document) is False
    key = list(sample_document.keys())[0]
    assert DocUtils().is_field("value", sample_document) is True
    assert DocUtils().is_field("blast32", sample_document) is True
    assert DocUtils().get_field("blast32", sample_document) == 21


def test_is_field_2(sample_document):
    assert DocUtils().is_field("blast32", sample_document) is True


@pytest.mark.skip(reason="Hell idk - need to think how to solve this problem")
def test_get_field(sample_document):
    assert (
        DocUtils().get_field("value.32", sample_document) == sample_document["value.32"]
    )
    assert DocUtils().get_field("value", sample_document) == sample_document["value"]


def test_get_field_across_documents(sample_document, sample_2_document):
    """Test to ensure you can get field across documents"""
    sample_docs = [sample_document] * 20 + [sample_2_document] * 100
    new_docs = DocUtils().get_field_across_documents(
        "value", sample_docs, missing_treatment="skip"
    )
    assert len(new_docs) == 20, "Not skipping"


def test_get_field_across_documents_for_skip_if_any_missing(
    sample_document, sample_2_document, combined_sample_document
):
    """Test to ensure you can get field across documents"""
    sample_docs = (
        [sample_document] * 20
        + [sample_2_document] * 100
        + [combined_sample_document] * 5
    )
    new_docs = DocUtils().get_fields_across_documents(
        ["value", "check_value"], sample_docs, missing_treatment="skip_if_any_missing"
    )
    assert len(new_docs) == 5, "Not skipping"


def test_subset_documents(combined_sample_document):
    sample_docs = [combined_sample_document] * 100
    subset_documents = DocUtils().subset_documents(
        ["value", "check_value"], sample_docs
    )
    for subset_doc in subset_documents:
        assert len(subset_doc) == 2
    assert len(subset_documents) == 100


def test_missing_treatment(combined_sample_document):
    assert -1 == DocUtils().get_field(
        "anonymous", combined_sample_document, missing_treatment=-1
    )


def test_inplace(sample_document):
    sample_document = sample_document

    new_sample_document = DocUtils().set_field(
        "blast32", sample_document, 22, inplace=False
    )

    assert sample_document["blast32"] == 21
    assert new_sample_document["blast32"] == 22



"""
    Testing dataset read operations
"""

import os
import uuid
import time
import pytest

import numpy as np
import pandas as pd

from relevanceai import Client

from relevanceai.dataset import Dataset

from tests.globals.constants import NOT_IMPLEMENTED, generate_dataset_id


class TestDatasetExport:
    def test_to_csv(self, test_dataset: Dataset):
        fname = f"{str(uuid.uuid4())}.csv"
        test_dataset.to_csv(fname)
        df = pd.read_csv(fname)

        assert df.values.size > 0

        os.remove(fname)

    def test_to_dict(self, test_dataset: Dataset):
        data = test_dataset.to_dict("records")

        assert data[0]

    def test_to_dataset(self, test_client: Client, test_dataset: Dataset):
        child_dataset_id = generate_dataset_id()

        test_dataset.to_dataset(child_dataset_id=child_dataset_id)
        child_dataset = test_client.Dataset(child_dataset_id)

        test_ids = [document["_id"] for document in test_dataset.get_all_documents()]
        child_ids = [document["_id"] for document in child_dataset.get_all_documents()]

        assert all(_id in test_ids for _id in child_ids)

    def test_to_pandas_dataframe(self, test_dataset: Dataset):
        df = test_dataset.to_pandas_dataframe()

        dataframe_ids = df.index.values.tolist()
        test_ids = [document["_id"] for document in test_dataset.get_all_documents()]

        assert all(_id in test_ids for _id in dataframe_ids)


class TestDatasetImport:
    pass


class TestDatasetSeries:
    def test_list_aliases(self, test_dataset: Dataset):
        series = test_dataset["sample_1_label"]
        aliases = series.list_aliases()
        assert "div" in aliases

    def test_head(self, test_dataset: Dataset):
        series = test_dataset["sample_2_label"]
        aliases = series.list_aliases()
        assert "div" in aliases

    def test_sample(self, test_dataset: Dataset):
        sample_n = test_dataset[["sample_1_label", "sample_2_label"]].sample(n=10)
        assert len(sample_n) == 10
        assert len(sample_n[0].keys()) == 3

        sample_n = test_dataset["sample_1_label"].sample(n=10)
        assert len(sample_n) == 10

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_all(self):
        assert False

    def test_apply(self, test_dataset: Dataset):
        before = test_dataset["sample_1_value"].values
        test_dataset["sample_1_value"].apply(
            lambda x: x + 1, output_field="sample_1_value"
        )
        after = test_dataset["sample_1_value"].values
        assert ((before + 1) == after).mean()

    def test_bulk_apply(self, test_dataset: Dataset):
        def bulk_func(documents):
            for document in documents:
                document["sample_1_value"] += 1
            return documents

        before = test_dataset["sample_1_value"].values
        test_dataset["sample_1_value"].bulk_apply(bulk_func)
        time.sleep(2)
        after = test_dataset["sample_1_value"].values

        assert ((before + 1) == after).mean()

    def test_numpy(self, test_dataset: Dataset):
        series = test_dataset["sample_1_value"]
        numpy = series.numpy()
        assert isinstance(numpy, np.ndarray)

    def test_add(self, test_dataset: Dataset):
        series_1 = test_dataset["sample_1_value"]
        series_2 = test_dataset["sample_2_value"]
        series_3 = test_dataset["sample_1_label"]

        series_1 + series_2

        try:
            series_1 + series_3
        except ValueError:
            assert True

        try:
            series_3 + series_2
        except ValueError:
            assert True

    def test_value_counts(self, test_dataset: Dataset):
        series = test_dataset["sample_1_value"]
        value_counts = series.value_counts()
        assert "sample_1_value" in value_counts.columns
        assert value_counts.size > 0

    def test_repr_html(self, test_dataset: Dataset):
        series = test_dataset["sample_1_value"]
        repr_html = series._repr_html_()
        assert "<" in repr_html
        assert ">" in repr_html


class TestDatasetStats:
    def test_value_counts(self, test_dataset: Dataset):
        value_counts = test_dataset.value_counts(field="sample_1_label")
        assert isinstance(value_counts, pd.DataFrame)

    def test_describe(self, test_dataset: Dataset):
        describe = test_dataset.describe()
        assert isinstance(describe, pd.DataFrame)

    @pytest.mark.skip(reason="requires matplotlib not sure how to handle")
    def test_corr(self, test_dataset: Dataset):
        test_dataset.cluster(model="kmeans", vector_fields=["sample_1_vector_"])
        corr = test_dataset.corr(
            X="sample_1_value",
            Y="sample_2_value",
            vector_field="sample_1_vector_",
            alias="kmeans-8",
            show_plot=False,
        )
        assert corr is None

    def test_health(self, test_dataset: Dataset):
        dataframe_output = test_dataset.health(output_format="dataframe")
        assert isinstance(dataframe_output, pd.DataFrame)

        json_output = test_dataset.health(output_format="json")
        assert isinstance(json_output, dict)

    @pytest.mark.skip(reason="Not sure why its failing")
    def test_aggregate(self, test_dataset: Dataset):
        agg = test_dataset.aggregate(groupby=["sample_1_label"])
        assert True

    @pytest.mark.skip(reason="Not sure why its failing")
    def test_facets(self, test_dataset: Dataset):
        facets = test_dataset.facets()
        assert facets


class TestDatasetMetadata:
    payload = {"test_metavalue": 1}

    def test_insert_metadata(self, test_dataset: Dataset):
        result = test_dataset.insert_metadata(self.payload)
        assert result is None

    def test_upsert_metadata(self, test_dataset: Dataset):
        result = test_dataset.upsert_metadata(self.payload)
        assert result is None

    def test_to_dict(self, test_dataset: Dataset):
        metadata = test_dataset.metadata.to_dict()
        assert metadata == {}

        result = test_dataset.insert_metadata(self.payload)
        assert result is None

        metadata = test_dataset.metadata.to_dict()
        assert metadata == self.payload


class TestDatasetRead:
    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_shape(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_info(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_head(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_get(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_schema(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_columns(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_filter(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_list_vector_fields(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_list_cluster_aliases(self):
        assert False


class TestDatasetWrite:
    def test_apply(self, test_dataset: Dataset):
        random_string = "you are the kingj"
        test_dataset["sample_1_label"].apply(
            lambda x: x + random_string, output_field="sample_1_label_2"
        )
        filtered_documents = test_dataset.datasets.documents.get_where(
            test_dataset.dataset_id,
            filters=[
                {
                    "field": "sample_1_label_2",
                    "filter_type": "contains",
                    "condition": "==",
                    "condition_value": random_string,
                }
            ],
        )
        assert len(filtered_documents["documents"]) > 0

    def test_bulk_apply(self, test_dataset: Dataset):
        random_string = "you are the queen"
        label = "sample_output"

        def bulk_fn(docs):
            for d in docs:
                d[label] = d.get("sample_1_label", "") + random_string
            return docs

        test_dataset.bulk_apply(bulk_fn)
        time.sleep(2)
        filtered_documents = test_dataset.datasets.documents.get_where(
            test_dataset.dataset_id,
            filters=[
                {
                    "field": "sample_output",
                    "filter_type": "contains",
                    "condition": "==",
                    "condition_value": random_string,
                }
            ],
        )
        assert len(filtered_documents["documents"]) > 0

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_concat(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_create(self):
        assert False

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_delete(self):
        assert False

    def test_insert_dataframe(self, test_dataset: Dataset):
        pandas_df = pd.DataFrame({"pandas_value": [3, 2, 1], "_id": ["10", "11", "12"]})
        test_dataset.insert_pandas_dataframe(pandas_df)
        assert "pandas_value" in pandas_df.columns

    @pytest.mark.skip(reason=NOT_IMPLEMENTED)
    def test_insert_csv(self):
        assert False

    @pytest.mark.skip(reason="Not used")
    def test_dtype_mapping(self, test_dataset: Dataset):
        test_dataset.set_dtypes({})
        dtypes = test_dataset.get_dtypes()
        assert all(
            dtype in dtypes
            for dtype in [
                "_numeric_",
                "_category_",
                "_text_",
                "_image_",
            ]
        )

    def test_insert_documents_with_media(self, test_dataset: Dataset):
        image_paths = ["tests/logo.png"] * 3
        pandas_df = pd.DataFrame({"image": image_paths, "_id": ["10", "11", "12"]})
        documents = pandas_df.to_dict("records")
        test_dataset.insert_documents(documents, media_fields=["image"])



from relevanceai.operations.operations import Operations
from relevanceai.operations.base import BaseOps
from relevanceai.operations.cluster import ClusterOps
from relevanceai.operations.cluster.sub import SubClusterOps
from relevanceai.operations.labels import LabelOps



from relevanceai.constants.errors import MissingPackageError
from relevanceai.operations_new.transform_base import TransformBase

try:
    from flair.data import Sentence
    from flair.models import SequenceTagger
except ModuleNotFoundError:
    raise MissingPackageError("flair", version="0.11.1")

from tqdm.auto import tqdm


class ExtractNounsTransform(TransformBase):
    """
    An operation for extracting noun
    """

    def __init__(
        self,
        fields: list,
        model_name: str,
        output_fields: list,
        cutoff_probability: float,
        stopwords: list = None,
        **kwargs
    ):
        self.fields = fields
        self.model_name = model_name
        self.output_fields = output_fields
        self.cutoff_probability = cutoff_probability
        self.tagger = SequenceTagger.load(model_name)
        from relevanceai.constants import STOPWORDS

        if stopwords is not None:
            self.stopwords = STOPWORDS + stopwords
        else:
            self.stopwords = STOPWORDS

        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def process_noun(noun):
        return (
            noun.replace("the ", "")
            .replace("this", "")
            .replace("'s ", "")
            .strip()
            .lower()
        )

    def extract_nouns(self, text, as_documents=False):
        sentence = Sentence(text)
        self.tagger.predict(sentence)
        nouns = []
        # Extract the noun phrases
        for entity in sentence.get_spans("np"):
            text = ExtractNounsTransform.process_noun(entity.text.lower().strip())

            if (
                entity.tag == "NP"
                and entity.score >= self.cutoff_probability
                and text not in self.stopwords
            ):
                if as_documents:
                    nouns.append({"noun": text, "score": entity.score})
                else:
                    nouns.append(text)
        return nouns

    def transform(self, docs):
        new_docs = [{"_id": d["_id"]} for d in docs]
        for i, d in enumerate(tqdm(docs)):
            for j, t in enumerate(self.fields):
                value = self.extract_nouns(
                    self.get_field(t, d, missing_treatment="return_empty_string")[:200],
                )
                self.set_field(self.output_fields[j], new_docs[i], value)
        return new_docs

    @property
    def name(self):
        return "extract-nouns"



from relevanceai.apps.report_app.sections import ReportSections


class ReportApp(ReportSections):
    def append_report(self, report):
        self.config["page-content"]["content"] = self.config["page-content"]["content"] + report.config["page-content"]["content"]

    def prepend_report(self, report):
        self.config["page-content"]["content"] = report.config["page-content"]["content"] + self.config["page-content"]["content"]



# -*- coding: utf-8 -*-
"""
Statistics API
"""
import pandas as pd

from typing import Optional, Union

from relevanceai.client.helpers import Credentials
from relevanceai.dataset.series import Series
from relevanceai.utils.decorators.version import added
from relevanceai.utils.decorators.analytics import track
from relevanceai._api import APIClient
from relevanceai._api.endpoints.datasets.cluster import ClusterClient


class Statistics(APIClient):
    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
        **kwargs,
    ):
        self.credentials = credentials
        self.dataset_id = dataset_id
        super().__init__(
            dataset_id=self.dataset_id,
            credentials=credentials,
            **kwargs,
        )

    @track
    def value_counts(self, field: str):
        """
        Return a Series containing counts of unique values.

        Parameters
        ----------
        field: str
            dataset field to which to do value counts on

        Returns
        -------
        Series

        Example
        -----------------
        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            field = "sample_field"
            value_counts_df = df.value_counts(field)

        """
        return Series(
            dataset=self,
            field=field,
        ).value_counts()

    @track
    def describe(self, return_type="pandas") -> dict:
        """
        Descriptive statistics include those that summarize the central tendency
        dispersion and shape of a dataset's distribution, excluding NaN values.


        Example
        -----------------
        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            field = "sample_field"
            df.describe() # returns pandas dataframe of stats
            df.describe(return_type='dict') # return raw json stats

        """
        facets = self.datasets.facets(self.dataset_id)
        if return_type == "pandas":
            schema = self.datasets.schema(self.dataset_id)
            dataframe = {
                col: facets["results"][col]
                for col in schema
                if col in facets["results"] and isinstance(facets["results"][col], dict)
            }
            dataframe = pd.DataFrame(dataframe)
            return dataframe
        elif return_type == "dict":
            return facets
        else:
            raise ValueError("invalid return_type, should be `dict` or `pandas`")

    @added(version="1.2.2")
    def corr(
        self,
        X: str,
        Y: str,
        vector_field: str,
        alias: str,
        groupby: Optional[str] = None,
        fontsize: int = 16,
        show_plot: bool = True,
    ):
        """
        Returns the Pearson correlation between two fields.

        Parameters
        ----------
        X: str
            A dataset field

        Y: str
            The other dataset field

        vector_field: str
            The vector field over which the clustering has been performed

        alias: str
            The alias of the clustering

        groupby: Optional[str]
            A field to group the correlations over

        fontsize: int
            The font size of the values in the image
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        # todo: how to cover cases when fields are in schema but not "calculable" fields like clusters and deployables
        cclient = ClusterClient(self.credentials)
        groupby_agg = (
            []
            if groupby is None
            else [{"name": groupby, "field": groupby, "agg": "correlation"}]
        )
        res = cclient.aggregate(
            dataset_id=self.dataset_id,
            vector_fields=[vector_field],
            metrics=[{"name": "correlation", "fields": [X, Y], "agg": "correlation"}],
            groupby=groupby_agg,
            alias=alias,
        )["results"]

        clusters = sorted(res.keys())

        if groupby is None:
            categories = ["cluster"]
        else:
            series = Series(
                dataset=self,
                field=groupby,
            ).all(show_progress_bar=False)

            categories = sorted(
                pd.Series(map(lambda _: _[groupby], series)).drop_duplicates()
            )

        # Use a pandas DataFrame for easy indexing
        dataframe = pd.DataFrame(data=[], columns=clusters, index=categories)

        for cluster, values in res.items():
            for value in values:
                correlation_value = value["correlation"][X][Y]
                category = value.get(groupby, "cluster")
                dataframe.at[category, cluster] = correlation_value

        # Only needed pandas DataFrame for indexing, now convert to numpy
        # ndarray for convenience.
        data = dataframe.to_numpy(dtype=float)

        fig, ax = plt.subplots(figsize=(2 * len(clusters), 2 * len(categories)))
        cmap = plt.get_cmap("coolwarm_r")
        cmap.set_bad("k")  # if np.nan, set imshow cell to black
        im = ax.imshow(data, norm=mpl.colors.Normalize(vmin=-1, vmax=1), cmap=cmap)

        ax.set_xticks(
            range(data.shape[1]),
            labels=clusters,
            rotation=-30,
            rotation_mode="anchor",
            ha="right",
            fontsize=fontsize + 1,
        )
        ax.set_yticks(range(data.shape[0]), labels=categories, fontsize=fontsize + 1)
        ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)

        ax.spines[:].set_visible(False)
        ax.set_xticks([n - 0.5 for n in range(data.shape[1] + 1)], minor=True)
        ax.set_yticks([n - 0.5 for n in range(data.shape[0] + 1)], minor=True)
        ax.grid(which="minor", color="w", linestyle="-", linewidth=3)
        ax.tick_params(which="minor", bottom=False, left=False)

        for column, _ in enumerate(clusters):
            for row, _ in enumerate(categories):
                # Ensure that the negative sign doesn't offset the text by
                # offseting positive values.
                if data[row][column] < 0:
                    text = f"{data[row][column]:.2f}"
                else:
                    text = f" {data[row][column]:.2f}"
                im.axes.text(
                    column,
                    row,
                    text,
                    dict(horizontalalignment="center", verticalalignment="center"),
                    fontsize=fontsize,
                )

        if show_plot:
            fig.tight_layout()
            plt.show()
        else:
            return fig

    def health(self, output_format="dataframe") -> Union[pd.DataFrame, dict]:
        """
        Gives you a summary of the health of your vectors, e.g. how many documents with vectors are missing, how many documents with zero vectors

        Parameters
        ----------

        output_format: str
            The format of the output. Must either be "dataframe" or "json".

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")
            df.health

        """
        print(
            "To see all rows, run `pd.set_option('display.max_rows', None)` (where `pd` is pandas)."
        )
        results = self.datasets.monitor.health(self.dataset_id)
        if output_format == "dataframe":
            return pd.DataFrame(results).T
        elif output_format == "json":
            return results
        else:
            raise ValueError('\'output_format\' must either be "dataframe" or "json"')

    @track
    def aggregate(
        self,
        groupby: Optional[list] = None,
        metrics: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        aggregation_query: Optional[dict] = None,
        sort: list = None,
    ):
        """
        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:

        Example
        ---------

        .. code-block::

            {
                "groupby" : [
                    {"name": <alias>, "field": <field in the collection>, "agg": "category"},
                    {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
                ],
                "metrics" : [
                    {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
                    {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
                ]
            }
            For example, one can use the following aggregations to group score based on region and player name.
            {
                "groupby" : [
                    {"name": "region", "field": "player_region", "agg": "category"},
                    {"name": "player_name", "field": "name", "agg": "category"}
                ],
                "metrics" : [
                    {"name": "average_score", "field": "final_score", "agg": "avg"},
                    {"name": "max_score", "field": "final_score", "agg": "max"},
                    {'name':'total_score','field':"final_score", 'agg':'sum'},
                    {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
                    {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
                ]
            }

        """
        return self.datasets.aggregate(
            dataset_id=self.dataset_id,
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            asc=asc,
            filters=filters,
            page_size=page_size,
            page=page,
            aggregation_query=aggregation_query,
        )

    @track
    def facets(
        self,
        fields: Optional[list] = [],
        date_interval: str = "monthly",
        page_size: int = 5,
        page: int = 1,
        asc: bool = False,
    ):
        """

        Get a summary of fields - such as most common, their min/max, etc.

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            from relevanceai.datasets import mock_documents
            documents = mock_documents(100)
            ds = client.Dataset("mock_documents")
            ds.upsert_documents(documents)
            ds.facets(["sample_1_value"])
        """
        return self.datasets.facets(
            dataset_id=self.dataset_id,
            fields=fields,
            date_interval=date_interval,
            page_size=page_size,
            page=page,
            asc=asc,
        )

    @track
    def health_check(self, **kwargs):
        # This is supposed to provide utilities
        raise NotImplementedError()
        # details = self.datasets.monitor.health(self.dataset_id, **kwargs)
        # return details



from typing import Optional

from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.dr.transform import DimReductionTransform


class DimReductionOps(DimReductionTransform, OperationAPIBase):
    """
    API related Functionality for Operation
    """

    def __init__(
        self,
        alias,
        model,
        n_components: int,
        model_kwargs: Optional[dict] = None,
        output_field: str = None,
        **kwargs
    ):
        if model_kwargs is None:
            model_kwargs = {}

        super().__init__(
            model=model,
            n_components=n_components,
            model_kwargs=model_kwargs,
            alias=alias,
            output_field=output_field,
            **kwargs,
        )
        self.model = self._get_model(
            model=model,
            n_components=n_components,
            alias=alias,
            **model_kwargs,
        )



"""

"""
from tqdm.auto import tqdm
from typing import Optional
from relevanceai.operations_new.cluster.text.explainer.base import BaseExplainer
from relevanceai.operations_new.ops_base import OperationAPIBase


class TextClusterExplainerOps(BaseExplainer, OperationAPIBase):  # type: ignore
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def transform(self, *args, **kwargs):
        # We transform isn't very useful here - it's mainly used for explain_clusters
        return self.explain_clusters(*args, **kwargs)

    def explain_clusters(
        self,
        dataset_id: str,
        vector_fields: list,
        alias: str,
        text_field,
        encode_fn,
        n_closest: int = 5,
        highlight_output_field="_explain_",
    ):

        """
        For each cluster, we explain the closest documents to the centroid, and then highlight the most
        important words based on what's most similar to the cluster center

        Parameters
        ----------
        text_field
            The field that contains the text to explain
        encode_fn
            The function that encodes the text. This is the same function that you used to encode the text in
        the dataset.
        n_closest : int, optional
            The number of closest documents to the centroid to explain
        highlight_output_field, optional
            The field that will be added to the document with the highlighted text.

        .. code-block::

            from relevanceai import Client
            client = Client()


        """
        # TODO:
        # Make this supported with cluster_ops.explain_text_clusters()

        # For each centroid, explains the closest n_closest to the centroids
        # Then highlights why the centroid is similar to the second most similar one
        # List the closest
        if len(vector_fields) > 1:
            raise NotImplementedError(
                "Currently do not support more than 1 vector fields."
            )

        from relevanceai.operations_new.cluster.ops import ClusterOps

        cluster_ops = ClusterOps(
            credentials=self.credentials,
            vector_fields=vector_fields,
            alias=alias,
            dataset_id=dataset_id,
            model=None,
        )
        # get get the centroid from the centroid
        closest = cluster_ops.list_closest(
            page_size=n_closest, select_fields=[text_field], verbose=False
        )
        # For the results in closest, we explain a text field
        # Flatten the closest
        for cluster, results in tqdm(closest["results"].items()):
            centroid_document = cluster_ops.get_centroid_from_id(cluster)
            query_vector = centroid_document[vector_fields[0]]

            if len(results["results"]) < 1:
                continue

            result_texts = self.get_field_across_documents(
                text_field, results["results"]
            )
            # Explain the first one
            for i, r in enumerate(result_texts):
                explained_answer = self.explain_from_vector(
                    encode_fn, query_vector=query_vector, answer_text=r
                )
                results["results"][i][highlight_output_field] = {
                    text_field: explained_answer
                }

            # Update explained results
            updated_results = self._update_documents(
                dataset_id=cluster_ops.dataset_id, documents=results["results"]
            )

        print(
            f"Make sure to set the highlight field `{text_field}` with substring `{highlight_output_field + '.' + text_field}` at: "
        )
        print(
            f"https://cloud.relevance.ai/dataset/{cluster_ops.dataset_id}/dashboard/settings"
        )
        return closest

    @property
    def name(self):
        return "text-explainer"

    def run(self):
        pass

    def explain_clusters_relational(
        self,
        dataset_id: str,
        vector_fields: list,
        alias: str,
        text_field,
        encode_fn,
        n_closest: int = 5,
        highlight_output_field="_explain_",
    ):

        """
        For each cluster, we explain the closest documents to the centroid, and then highlight the most
        important words in the centroid and then for the centroid, we analyse what makes it similar to the second most similar document

        Parameters
        ----------
        cluster_ops : ClusterOps
            ClusterOps,
        text_field
            The field that contains the text to explain
        encode_fn
            The function that encodes the text. This is the same function that you used to encode the text in
        the dataset.
        n_closest : int, optional
            The number of closest documents to the centroid to explain
        highlight_output_field, optional
            The field that will be added to the document with the highlighted text.
        .. code-block::
            from relevanceai import Client
            client = Client()
        """
        # TODO:
        # Make this supported with cluster_ops.explain_text_clusters()

        # For each centroid, explains the closest n_closest to the centroids
        # Then highlights why the centroid is similar to the second most similar one
        # List the closest

        if len(vector_fields) > 1:
            raise NotImplementedError(
                "Currently do not support more than 1 vector fields."
            )

        from relevanceai.operations_new.cluster.ops import ClusterOps

        cluster_ops = ClusterOps(
            credentials=self.credentials,
            vector_fields=vector_fields,
            alias=alias,
            dataset_id=dataset_id,
            model=None,
        )

        closest = cluster_ops.list_closest(
            page_size=n_closest, select_fields=[text_field], verbose=False
        )
        # For the results in closest, we explain a text field
        # Flatten the closest
        for cluster, results in tqdm(closest["results"].items()):
            if len(results["results"]) < 2:
                continue
            query_text = self.get_field(text_field, results["results"][0])
            result_texts = self.get_field_across_documents(
                text_field, results["results"][1:]
            )
            # Explain the first one
            explained_answer = self.explain(
                encode_fn, query_text=result_texts[0], answer_text=query_text
            )

            # results["results"][0][highlight_output_field] = {
            #     text_field: explained_answer
            # }
            self.set_field(
                highlight_output_field + "." + text_field,
                results["results"][0],
                explained_answer,
            )
            for i, r in enumerate(result_texts):
                explained_answer = self.explain(
                    encode_fn, query_text=query_text, answer_text=r
                )
                split_text_fields = text_field.split(".")
                # Here we want to store nested fields
                # For example "value.check": "this is answer"
                # should be stored like this:
                # {"value": {"check": "this is answer"}}
                self.set_field(
                    f"{highlight_output_field}.{text_field}",
                    results["results"][i + 1],
                    explained_answer,
                )

            # Update explained results
            updated = self._update_documents(
                dataset_id=cluster_ops.dataset_id, documents=results["results"]
            )

        print(
            f"Make sure to set the highlight field `{text_field}` with substring `{highlight_output_field + '.' + text_field}` at: "
        )
        print(
            f"https://cloud.relevance.ai/dataset/{cluster_ops.dataset_id}/dashboard/settings"
        )
        self.datasets.post_settings(
            dataset_id=cluster_ops.dataset_id,
            settings={
                "settings": {
                    "highlightingRules": [
                        {
                            "substringField": highlight_output_field,
                            "weightField": "",
                            "fullField": text_field,
                        }
                    ]
                }
            },
        )
        return closest



"""Add Sentiment to your dataset
"""

# Running a function across each subcluster
import pandas as pd
import numpy as np
import csv
from typing import Dict, List, Optional
from urllib.request import urlopen
from relevanceai.constants.errors import MissingPackageError
from relevanceai.operations_new.transform_base import TransformBase


class SentimentTransform(TransformBase):
    def __init__(
        self,
        text_fields: list,
        model_name: str = "cardiffnlp/twitter-roberta-base-sentiment",
        highlight: bool = False,
        positive_sentiment_name: str = "positive",
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
        output_fields: list = None,
        sensitivity: float = 0,
        device: int = None,
        strategy: str = "value_max",
        eps: float = 1e-9,
        **kwargs,
    ):
        """
        Sentiment Ops.

        Parameters
        -------------

        model_name: str
            The name of the model
        sensitivity: float
            How confident it is about being `neutral`. If you are dealing with news sources,
            you probably want less sensitivity

        """
        self.model_name = model_name
        self.text_fields = text_fields
        self.highlight = highlight
        self.positive_sentiment_name = positive_sentiment_name
        self.max_number_of_shap_documents = max_number_of_shap_documents
        self.min_abs_score = min_abs_score
        self.output_fields = (
            output_fields
            if output_fields is not None
            else [self._get_output_field(t) for t in text_fields]
        )
        self.sensitivity = sensitivity
        self.strategy = strategy
        self.eps = eps
        self.device = self.get_transformers_device(device)

        import transformers

        # Tries to set it to set it on GPU
        self._classifier = transformers.pipeline(
            return_all_scores=True, model=self.model_name, device=self.device
        )

        for k, v in kwargs.items():
            setattr(self, k, v)

    def preprocess(self, text: str):
        new_text = []
        for t in text.split(" "):
            t = "@user" if t.startswith("@") and len(t) > 1 else t
            t = "http" if t.startswith("http") else t
            new_text.append(t)
        return " ".join(new_text)

    @property
    def classifier(self):
        return self._classifier

    def _get_label_mapping(self, task: str):
        # Note: this is specific to the current model
        labels = []
        mapping_link = f"https://raw.githubusercontent.com/cardiffnlp/tweeteval/main/datasets/{task}/mapping.txt"
        with urlopen(mapping_link) as f:
            html = f.read().decode("utf-8").split("\n")
            csvreader = csv.reader(html, delimiter="\t")
        labels = [row[1] for row in csvreader if len(row) > 1]
        return labels

    @property
    def label_mapping(self):
        return {"LABEL_0": "negative", "LABEL_1": "neutral", "LABEL_2": "positive"}

    def _get_scores(
        self, cls: str, logits: List[List[Dict[str, float]]]
    ) -> List[float]:
        return [
            list(filter(lambda label: label["label"] == cls, logit))[0]["score"]
            for logit in logits
        ]

    def analyze_sentiment(
        self,
        texts: List[str],
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
    ):
        logits = self.classifier(texts, truncation=True, max_length=512)
        scores = pd.DataFrame(
            {
                "negative": self._get_scores(cls="LABEL_0", logits=logits),
                "neutral": self._get_scores(cls="LABEL_1", logits=logits),
                "positive": self._get_scores(cls="LABEL_2", logits=logits),
            }
        )

        if not self.highlight:
            if self.strategy == "expected_value":
                scores, labels = self._expected_value_scoring(
                    scores=scores,
                    texts=texts,
                )

            elif self.strategy == "value_max":
                scores, labels = self._value_max_scoring(
                    scores=scores,
                    texts=texts,
                )

            else:
                raise ValueError

            updates = [
                {
                    "sentiment": scores[index],
                    "overall_sentiment_score": labels[index],
                }
                for index in range(len(texts))
            ]
            return updates

        else:
            return self._shap_documents(
                scores=scores,
                texts=texts,
                max_number_of_shap_documents=max_number_of_shap_documents,
                min_abs_score=min_abs_score,
            )

    def _get_idxmax(self, scores: pd.DataFrame, texts: List[str]):
        return [scores.iloc[index].idxmax(-1) for index in range(len(texts))]

    def _get_value_max_scores(self, scores: pd.DataFrame):
        negative_mask = scores["negative"].values > scores["positive"].values
        positive_mask = scores["negative"].values <= scores["positive"].values
        a_max = (
            -scores["negative"].values * negative_mask
            + scores["positive"].values * positive_mask
        )
        a_max[a_max == 0] += self.eps
        return a_max

    def _value_max_scoring(
        self,
        scores: pd.DataFrame,
        texts: List[str],
    ):
        a_max = self._get_value_max_scores(scores=scores)
        idxmax = self._get_idxmax(scores=scores, texts=texts)
        return a_max, idxmax

    def _get_expected_value_score(self, scores: pd.DataFrame):
        # this is the expected value
        e_x = scores["negative"].values * -1
        e_x += scores["positive"].values
        e_x[e_x == 0] += self.eps
        return e_x

    def _expected_value_scoring(
        self,
        scores: pd.DataFrame,
        texts: List[str],
    ):
        e_x = self._get_expected_value_score(scores=scores)
        idxmax = self._get_idxmax(scores=scores, texts=texts)
        return e_x, idxmax

    def _shap_documents(
        self,
        scores: pd.DataFrame,
        texts,
        max_number_of_shap_documents: int = None,
        min_abs_score: float = 0.1,
    ):
        idxmax = self._get_idxmax(scores=scores, texts=texts)
        shap_documents = [
            self.get_shap_values(
                texts[index],
                sentiment_ind=idxmax[index],
                max_number_of_shap_documents=max_number_of_shap_documents,
                min_abs_score=min_abs_score,
            )
            for index in range(len(texts))
        ]

        max_scores = scores.values[np.argmax(scores.values, axis=-1)]
        e_x = self._get_expected_value_score(scores=scores)

        updates = [
            {
                "sentiment": idxmax[index],
                "score": max_scores[index],
                "overall_sentiment": e_x[index],
                "highlight_chunk_": shap_documents[index],
            }
            for index in range(len(texts))
        ]
        return updates

    @property
    def explainer(self):
        if hasattr(self, "_explainer"):
            return self._explainer
        else:
            try:
                import shap
            except ModuleNotFoundError:
                raise MissingPackageError("shap")
            self._explainer = shap.Explainer(self.classifier)
            return self._explainer

    def _calculate_overall_sentiment(self, score: float, label: str) -> float:
        if label.lower().strip() == "positive":
            return score
        else:
            return -score

    def get_shap_values(
        self,
        text: str,
        sentiment_ind: int = 2,
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
    ):
        """Get SHAP values"""
        shap_values = self.explainer([text])
        cohorts = {"": shap_values}
        cohort_labels = list(cohorts.keys())
        cohort_exps = list(cohorts.values())
        features = cohort_exps[0].data
        feature_names = cohort_exps[0].feature_names
        values = np.array([cohort_exps[i].values for i in range(len(cohort_exps))])
        shap_docs = [
            {"text": v, "score": f}
            for f, v in zip(
                [x[sentiment_ind] for x in values[0][0].tolist()], feature_names[0]
            )
        ]
        if max_number_of_shap_documents is not None:
            sorted_scores = sorted(shap_docs, key=lambda x: x["score"], reverse=True)
        else:
            sorted_scores = sorted(shap_docs, key=lambda x: x["score"], reverse=True)[
                :max_number_of_shap_documents
            ]
        return [d for d in sorted_scores if abs(d["score"]) > min_abs_score]

    @property
    def name(self):
        return "sentiment"

    def _get_output_field(self, text_field):
        norm_name = self.model_name.replace("/", "-")
        return f"_sentiment_.{text_field}.{norm_name}"

    def transform(self, documents):
        # For each document, update the field
        sentiment_docs = [{"_id": d["_id"]} for d in documents]
        for i, t in enumerate(self.text_fields):
            sentiments = self.analyze_sentiment(
                [
                    self.get_field(t, doc, missing_treatment="return_empty_string")
                    for doc in documents
                ],
                max_number_of_shap_documents=self.max_number_of_shap_documents,
                min_abs_score=self.min_abs_score,
            )
            output_field = self.output_fields[i]
            self.set_field_across_documents(output_field, sentiments, sentiment_docs)
        return sentiment_docs



import getpass

from dataclasses import dataclass
from typing import List

from relevanceai.constants.messages import Messages
from relevanceai.constants import (
    SIGNUP_URL,
    AUSTRALIA_URL,
    WIDER_URL,
    OLD_AUSTRALIA_EAST,
    DEV_URL,
)
from relevanceai.constants.errors import (
    APIKeyNotFoundError,
    FireBaseUIDNotFoundError,
    ProjectNotFoundError,
    RegionNotFoundError,
    TokenNotFoundError,
)


def region_to_url(region: str) -> str:
    """A function to convert a users region to base url for API calls

    Args:
        region (str): user region, element of ["us-east-1", "ap-southeast-2", "old-australia-east"]

    Returns:
        url: the appropriate base url for API calls
    """
    if "dev" in region:
        actual_region = region.replace("dev-", "")
        url = DEV_URL.format(actual_region)

    elif region == OLD_AUSTRALIA_EAST:
        url = AUSTRALIA_URL

    else:
        url = WIDER_URL.format(region)

    return url


@dataclass
class Credentials:
    """
    A convenience store of relevant credentials.
    """

    __slots__ = (
        "token",
        "project",
        "api_key",
        "region",
        "firebase_uid",
    )
    token: str
    project: str
    api_key: str
    region: str
    firebase_uid: str

    def split_token(self) -> List[str]:
        """
        Splits the token in its four main components given in order by above

        Returns:
            List[str]: a list of strings containing all user identification for interaction with dashboard and api calls
        """
        return self.token.split(":")

    def dict(self) -> dict:
        return {
            "project": self.project,
            "api_key": self.api_key,
            "region": self.region,
            "firebase_uid": self.firebase_uid,
            "token": self.token,
        }


def process_token(token: str):
    """Given a user token, checks to see if all necessary credentials are present in token.

    Args:
        token (str): a ":" delimited string of identifying information

    Raises:
        TokenNotFoundError: error if idenitifier is not found
        ProjectNotFoundError: error if idenitifier is not found
        APIKeyNotFoundError: error if idenitifier is not found
        RegionNotFoundError: error if idenitifier is not found
        FireBaseUIDNotFoundError: error if idenitifier is not found

    Returns:
        Credentials: a dataclass of all user identification
    """
    if not token:
        raise TokenNotFoundError

    # A token takes the format project:api_key:region:firebase_uid
    project, api_key, *other_credentials = token.split(":")

    if not project:
        raise ProjectNotFoundError

    if not api_key:
        raise APIKeyNotFoundError

    if len(other_credentials) == 0:
        # Assume that region is missing as that is the next credential
        raise RegionNotFoundError
    elif len(other_credentials) == 1:
        # Assume region exists but firebase_uid, the next credential, does not
        raise FireBaseUIDNotFoundError
    else:
        # Two or more other credentials is correct
        region, firebase_uid, *additional_credentials = other_credentials

    return Credentials(token, project, api_key, region, firebase_uid)


def auth() -> str:
    """_summary_

    Returns:
        token: a ":" delimited string of identifying information
    """
    print(Messages.TOKEN_MESSAGE.format(SIGNUP_URL))
    token = getpass.getpass(f"Activation Token: ")
    return token



import random

import pytest

from relevanceai import Client
from relevanceai.dataset import Dataset
from relevanceai.utils.datasets import (
    get_iris_dataset,
    get_online_ecommerce_dataset,
    get_palmer_penguins_dataset,
)
from relevanceai.utils.decorators.vectors import catch_errors


from tests.globals.constants import (
    SAMPLE_DATASET_DATASET_PREFIX,
    generate_random_label,
    generate_random_string,
    generate_random_vector,
    generate_random_integer,
)


MULTIVECTOR_QUERY = [
    {"vector": generate_random_vector(), "fields": ["sample_1_vector_"]}
]

CHUNK_MULTIVECTOR_QUERY = [
    {"vector": generate_random_vector(), "fields": ["_chunk_.label_chunkvector_"]}
]


class TestVectorizeOps:
    @pytest.mark.xfail(reason="vectorhub is not required")
    def test_vectorize(self, test_client: Client):
        dataset = test_client.Dataset(SAMPLE_DATASET_DATASET_PREFIX + "_ecom")

        documents = get_online_ecommerce_dataset()
        dataset.insert_documents(documents=documents)

        dataset.vectorize(create_feature_vector=True)

        assert True

    def test_numeric_vectorize(self, test_client: Client):
        dataset = test_client.Dataset(SAMPLE_DATASET_DATASET_PREFIX + "_iris")

        documents = get_iris_dataset()
        dataset.insert_documents(documents=documents)

        dataset.vectorize(create_feature_vector=True)

        assert "_dim4_feature_vector_" in dataset.schema

        dataset.vectorize(fields=["numeric"], create_feature_vector=True)

        assert "_dim4_feature_vector_" in dataset.schema

    @pytest.mark.xfail(reason="deprecated")
    def test_custom_vectorize(self, test_client: Client):
        dataset = test_client.Dataset(SAMPLE_DATASET_DATASET_PREFIX + "_penguins")

        documents = get_palmer_penguins_dataset()
        dataset.insert_documents(documents=documents)

        from relevanceai.operations.vector import Base2Vec

        class CustomTextEncoder(Base2Vec):
            __name__ = "CustomTextEncoder".lower()

            def __init__(self, *args, **kwargs):
                super().__init__()

                self.vector_length = 128
                self.model = lambda x: [
                    random.random() if "None" not in str(x) else random.random() + 10
                    for _ in range(self.vector_length)
                ]

            @catch_errors
            def encode(self, value):
                vector = self.model(value)
                return vector

        dataset.vectorize(
            encoders=dict(
                text=[
                    CustomTextEncoder(),
                ],
            ),
            create_feature_vector=True,
        )

        vectors = [
            "Comments_customtextencoder_vector_",
            "Species_customtextencoder_vector_",
            "Stage_customtextencoder_vector_",
        ]
        assert all(vector in dataset.schema for vector in vectors)

    # def test_vector_search(self, test_dataset: Dataset):
    #     test_dataset.vector_search(
    #         multivector_query=MULTIVECTOR_QUERY,
    #     )
    #     assert True

    # def test_hybrid_search(self, test_dataset: Dataset):
    #     test_dataset.hybrid_search(
    #         multivector_query=MULTIVECTOR_QUERY,
    #         text="hey",
    #         fields=["sample_1_label"],
    #     )
    #     assert True

    # def test_chunk_search(self, test_dataset: Dataset):
    #     test_dataset.chunk_search(
    #         multivector_query=CHUNK_MULTIVECTOR_QUERY,
    #         chunk_field="_chunk_",
    #     )
    #     assert True

    # def test_multistep_chunk_search(self, test_dataset: Dataset):
    #     test_dataset.multistep_chunk_search(
    #         multivector_query=CHUNK_MULTIVECTOR_QUERY,
    #         first_step_multivector_query=MULTIVECTOR_QUERY,
    #         chunk_field="_chunk_",
    #     )
    #     assert True



"""
TransformBase
"""
from typing import Any, Dict, Optional, List
from relevanceai.operations_new.cluster.alias import ClusterAlias
from relevanceai.operations_new.cluster.batch.models.base import BatchClusterModelBase
from relevanceai.operations_new.transform_base import TransformBase
from relevanceai.operations_new.cluster.transform import ClusterTransform


class BatchClusterTransform(ClusterTransform, ClusterAlias):
    def __init__(
        self,
        vector_fields: list,
        model: Any,
        model_kwargs: dict,
        *args,
        **kwargs,
    ):
        self.vector_fields = vector_fields
        self._check_vector_fields()
        self.model = self._get_model(model, model_kwargs)
        if model_kwargs is None:
            model_kwargs = {}
        self.model_kwargs: dict = model_kwargs

        for k, v in kwargs.items():
            setattr(self, k, v)

    def partial_fit(self, documents: List, model_kwargs=None):
        """Run partial fitting on a list of documents"""
        if isinstance(self.model, str):
            model_kwargs = {} if model_kwargs is None else model_kwargs
            self.model = self._get_model(self.model, model_kwargs)
        return self.model.partial_fit(
            self.get_field_across_documents(self.vector_fields, documents)
        )

    @property
    def name(self):
        return type(self.model).__name__.lower()

    @property
    def full_cluster_field(self):
        if not hasattr(self, "_cluster_field"):
            self._cluster_field = self._get_cluster_field_name()
            print("Saving cluster labels to: ")
            print(self._cluster_field)
        return self._cluster_field

    def transform(self, documents: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """It takes a list of documents, and for each document, it runs the document through each of the
        models in the pipeline, and returns the updated documents.

        Parameters
        ----------
        documents : List[Dict[str, Any]]
            List[Dict[str, Any]]

        Returns
        -------
            A list of dictionaries.

        """

        # TODO: add support for sklearn kmeans
        labels = self.predict_documents(
            documents=documents,
        )
        # Get the cluster field name
        cluster_field_name = self._get_cluster_field_name()

        documents_to_upsert = [{"_id": d["_id"]} for d in documents]

        self.set_field_across_documents(
            cluster_field_name,
            labels,
            documents_to_upsert,
        )
        return documents_to_upsert

    def predict_documents(self, documents, warm_start=False):
        """
        If warm_start=True, copies the values from the previous fit.
        Only works for cluster models that use centroids. You should
        not have to use this parameter.
        """
        # run fit predict on documetns
        if hasattr(self.model, "predict_documents"):
            cluster_labels = self.model.predict_documents(
                documents=documents,
                vector_fields=self.vector_fields,
            )
            return self.format_cluster_labels(cluster_labels)
        elif hasattr(self.model, "predict"):
            if len(self.vector_fields) == 1:
                vectors = self.get_field_across_documents(
                    self.vector_fields[0],
                    documents,
                )
                cluster_labels = self.model.predict(vectors)

                return self.format_cluster_labels(cluster_labels)

        raise AttributeError(
            "Model is missing a `predict` or `predict_documents` method."
        )

    def _get_model(self, model: Any, model_kwargs: dict) -> Any:
        if isinstance(model, str):
            model = self._get_model_from_string(model, model_kwargs)

        elif "sklearn" in model.__module__:
            model = self._get_sklearn_model_from_class(model)

        elif "faiss" in model.__module__:
            model = self._get_faiss_model_from_class(model)

        return model



from typing import List, Union, Optional

import numpy as np

from relevanceai.operations_new.scaling.models.base import ScalerModelBase

from sklearn.base import TransformerMixin

from relevanceai.operations_new.scaling.models import sklearn_models


class SKLearnScaler(ScalerModelBase):
    def __init__(
        self,
        model: Union[str, TransformerMixin],
        alias: Optional[str] = None,
        model_kwargs: Optional[dict] = None,
    ):
        self.alias = alias
        self.model_kwargs = model_kwargs
        self.model: TransformerMixin = self._get_model(model)
        self.model_name = type(self.model).__name__.lower()

    def _get_model(self, model: Union[str, TransformerMixin]):

        if isinstance(model, str):
            model_string = sklearn_models[model]
            model = ScalerModelBase.import_from_string(
                f"sklearn.preprocessing.{model_string}"
            )

            model = model(**self.model_kwargs)

        return model

    def fit(
        self,
        vectors: Union[List[List[float]], np.ndarray],
    ) -> None:
        """It fits the model to the vectors.

        Parameters
        ----------
        vectors : Union[List[List[float]], np.ndarray]
            Union[List[List[float]], np.ndarray]

        """

        if isinstance(vectors, list):
            vectors = np.array(vectors)

        self.model.fit(vectors)

    def fit_transform(
        self,
        vectors: List[List[float]],
    ) -> List[List[float]]:
        """It takes a list of vectors, fits the model to the vectors, and then transforms the vectors

        Parameters
        ----------
        vectors : List[List[float]]
            List[List[float]]

        Returns
        -------
            A list of lists of floats.

        """

        vectors = np.array(vectors)
        reduced_vectors = self.model.fit_transform(vectors)
        return reduced_vectors.tolist()



"""Search In A Dataset
"""
from typing import List, Optional

from relevanceai.client.helpers import Credentials
from relevanceai.utils.decorators.analytics import track
from relevanceai.utils.decorators.version import deprecated_error
from relevanceai._api import APIClient


class SearchOps(APIClient):
    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
    ):
        self.dataset_id = dataset_id

        super().__init__(credentials)

    @track
    @deprecated_error("Please use the new `dataset.search` method.")
    def vector_search(
        self,
        multivector_query: List,
        positive_document_ids: Optional[dict] = None,
        negative_document_ids: Optional[dict] = None,
        vector_operation="sum",
        approximation_depth=0,
        sum_fields=True,
        page_size=20,
        page=1,
        similarity_metric="cosine",
        facets: Optional[list] = None,
        filters: Optional[list] = None,
        min_score=0,
        select_fields: Optional[list] = None,
        include_vector=False,
        include_count=True,
        asc=False,
        keep_search_history=False,
        hundred_scale=False,
        search_history_id=None,
        query: str = None,
    ):
        """
        Allows you to leverage vector similarity search to create a semantic search engine. Powerful features of Relevance vector search:

        1. Multivector search that allows you to search with multiple vectors and give each vector a different weight.
        e.g. Search with a product image vector and text description vector to find the most similar products by what it looks like and what its described to do.
        You can also give weightings of each vector field towards the search, e.g. image_vector_ weights 100%, whilst description_vector_ 50% \n
            An example of a simple multivector query:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": ["name_vector_"], "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": ["image_vector_"], "alias":"image"},
            >>> ]

            An example of a weighted multivector query:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": {"name_vector_":0.6}, "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": {"image_vector_"0.4}, "alias":"image"},
            >>> ]

            An example of a weighted multivector query with multiple fields for each vector:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": {"name_vector_":0.6, "description_vector_":0.3}, "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": {"image_vector_"0.4}, "alias":"image"},
            >>> ]

        2. Utilise faceted search with vector search. For information on how to apply facets/filters check out datasets.documents.get_where \n
        3. Sum Fields option to adjust whether you want multiple vectors to be combined in the scoring or compared in the scoring. e.g. image_vector_ + text_vector_ or image_vector_ vs text_vector_. \n
            When sum_fields=True:

            - Multi-vector search allows you to obtain search scores by taking the sum of these scores.
            - TextSearchScore + ImageSearchScore = SearchScore
            - We then rank by the new SearchScore, so for searching 1000 documents there will be 1000 search scores and results

            When sum_fields=False:

            - Multi vector search but not summing the score, instead including it in the comparison!
            - TextSearchScore = SearchScore1
            - ImagSearchScore = SearchScore2
            - We then rank by the 2 new SearchScore, so for searching 1000 documents there should be 2000 search scores and results.

        4. Personalization with positive and negative document ids.

            - For more information about the positive and negative document ids to personalize check out services.recommend.vector

        For more even more advanced configuration and customisation of vector search, reach out to us at dev@relevance.ai and learn about our new advanced_vector_search.

        Parameters
        ----------

        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: string
            Search history ID, only used for storing search histories.
        query: string
            What to store as the query name in the dashboard

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            results = df.vector_search(multivector_query=MULTIVECTOR_QUERY)

        """

        positive_document_ids = (
            {} if positive_document_ids is None else positive_document_ids
        )
        negative_document_ids = (
            {} if negative_document_ids is None else negative_document_ids
        )
        facets = [] if facets is None else facets
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        return self.services.search.vector(
            dataset_id=self.dataset_id,
            multivector_query=multivector_query,
            positive_document_ids=positive_document_ids,
            negative_document_ids=negative_document_ids,
            vector_operation=vector_operation,
            approximation_depth=approximation_depth,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            facets=facets,
            filters=filters,
            min_score=min_score,
            select_fields=select_fields,
            include_vector=include_vector,
            include_count=include_count,
            asc=asc,
            keep_search_history=keep_search_history,
            hundred_scale=hundred_scale,
            search_history_id=search_history_id,
            query=query,
        )

    @track
    @deprecated_error("Please use the new `dataset.search` method.")
    def hybrid_search(
        self,
        multivector_query: List,
        text: str,
        fields: list,
        edit_distance: int = -1,
        ignore_spaces: bool = True,
        traditional_weight: float = 0.075,
        page_size: int = 20,
        page=1,
        similarity_metric="cosine",
        facets: Optional[list] = None,
        filters: Optional[list] = None,
        min_score=0,
        select_fields: Optional[list] = None,
        include_vector=False,
        include_count=True,
        asc=False,
        keep_search_history=False,
        hundred_scale=False,
        search_history_id=None,
        sum_fields: bool = False,
    ):
        """
        Combine the best of both traditional keyword faceted search with semantic vector search to create the best search possible. \n

        For information on how to use vector search check out services.search.vector. \n

        For information on how to use traditional keyword faceted search check out services.search.traditional.

        Parameters
        ----------
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        text : string
            Text Search Query (not encoded as vector)
        fields : list
            Text fields to search against
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: float
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: string
            Search history ID, only used for storing search histories.
        edit_distance: int
            This refers to the amount of letters it takes to reach from 1 string to another string. e.g. band vs bant is a 1 word edit distance. Use -1 if you would like this to be automated.
        ignore_spaces: bool
            Whether to consider cases when there is a space in the word. E.g. Go Pro vs GoPro.
        traditional_weight: int
            Multiplier of traditional search score. A value of 0.025~0.075 is the ideal range


        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            MULTIVECTOR_QUERY = [{"vector": [0, 1, 2], "fields": ["sample_vector_"]}]
            results = df.vector_search(multivector_query=MULTIVECTOR_QUERY)

        """
        facets = [] if facets is None else facets
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        return self.services.search.hybrid(
            dataset_id=self.dataset_id,
            multivector_query=multivector_query,
            text=text,
            fields=fields,
            edit_distance=edit_distance,
            ignore_spaces=ignore_spaces,
            traditional_weight=traditional_weight,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            facets=facets,
            filters=filters,
            min_score=min_score,
            select_fields=select_fields,
            include_vector=include_vector,
            include_count=include_count,
            asc=asc,
            keep_search_history=keep_search_history,
            hundred_scale=hundred_scale,
            search_history_id=search_history_id,
            sum_fields=sum_fields,
        )

    @track
    @deprecated_error("Please use the new `dataset.search` method.")
    def chunk_search(
        self,
        multivector_query,
        chunk_field,
        chunk_scoring="max",
        chunk_page_size: int = 3,
        chunk_page: int = 1,
        approximation_depth: int = 0,
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        facets: Optional[list] = None,
        filters: Optional[list] = None,
        min_score: int = None,
        include_vector: bool = False,
        include_count: bool = True,
        asc: bool = False,
        keep_search_history: bool = False,
        hundred_scale: bool = False,
        query: str = None,
    ):
        """
        Chunks are data that has been divided into different units. e.g. A paragraph is made of many sentence chunks, a sentence is made of many word chunks, an image frame in a video. By searching through chunks you can pinpoint more specifically where a match is occuring. When creating a chunk in your document use the suffix "chunk" and "chunkvector". An example of a document with chunks:

        >>> {
        >>>     "_id" : "123",
        >>>     "title" : "Lorem Ipsum Article",
        >>>     "description" : "Lorem Ipsum is simply dummy text of the printing and typesetting industry. Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book. It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged.",
        >>>     "description_vector_" : [1.1, 1.2, 1.3],
        >>>     "description_sentence_chunk_" : [
        >>>         {"sentence_id" : 0, "sentence_chunkvector_" : [0.1, 0.2, 0.3], "sentence" : "Lorem Ipsum is simply dummy text of the printing and typesetting industry."},
        >>>         {"sentence_id" : 1, "sentence_chunkvector_" : [0.4, 0.5, 0.6], "sentence" : "Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book."},
        >>>         {"sentence_id" : 2, "sentence_chunkvector_" : [0.7, 0.8, 0.9], "sentence" : "It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged."},
        >>>     ]
        >>> }

        For combining chunk search with other search check out services.search.advanced_chunk.

        Parameters
        ----------

        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        chunk_field : string
            Field where the array of chunked documents are.
        chunk_scoring: string
            Scoring method for determining for ranking between document chunks.
        chunk_page_size: int
            Size of each page of chunk results
        chunk_page: int
            Page of the chunk results
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        query: string
            What to store as the query name in the dashboard

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            results = df.chunk_search(
                chunk_field="_chunk_",
                multivector_query=MULTIVECTOR_QUERY
            )

        """
        raise DeprecationWarning()  # Should error as above based on decorator

    @track
    @deprecated_error("Please use the new `dataset.search` method.")
    def multistep_chunk_search(
        self,
        multivector_query,
        first_step_multivector_query,
        chunk_field,
        chunk_scoring="max",
        chunk_page_size: int = 3,
        chunk_page: int = 1,
        approximation_depth: int = 0,
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        facets: Optional[list] = None,
        filters: Optional[list] = None,
        min_score: int = None,
        include_vector: bool = False,
        include_count: bool = True,
        asc: bool = False,
        keep_search_history: bool = False,
        hundred_scale: bool = False,
        first_step_page: int = 1,
        first_step_page_size: int = 20,
        query: str = None,
    ):
        """
        Multistep chunk search involves a vector search followed by chunk search, used to accelerate chunk searches or to identify context before delving into relevant chunks. e.g. Search against the paragraph vector first then sentence chunkvector after. \n

        For more information about chunk search check out datasets.search.chunk. \n

        For more information about vector search check out services.search.vector

        Parameters
        ----------

        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        chunk_field : string
            Field where the array of chunked documents are.
        chunk_scoring: string
            Scoring method for determining for ranking between document chunks.
        chunk_page_size: int
            Size of each page of chunk results
        chunk_page: int
            Page of the chunk results
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        first_step_multivector_query: list
            Query for advance search that allows for multiple vector and field querying.
        first_step_page: int
            Page of the results
        first_step_page_size: int
            Size of each page of results
        query: string
            What to store as the query name in the dashboard

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            results = df.search.multistep_chunk(
                chunk_field="_chunk_",
                multivector_query=MULTIVECTOR_QUERY,
                first_step_multivector_query=FIRST_STEP_MULTIVECTOR_QUERY
            )

        """
        facets = [] if facets is None else facets
        filters = [] if filters is None else filters

        return self.services.search.multistep_chunk(
            dataset_id=self.dataset_id,
            multivector_query=multivector_query,
            first_step_multivector_query=first_step_multivector_query,
            chunk_field=chunk_field,
            chunk_scoring=chunk_scoring,
            chunk_page_size=chunk_page_size,
            chunk_page=chunk_page,
            approximation_depth=approximation_depth,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            facets=facets,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            asc=asc,
            keep_search_history=keep_search_history,
            hundred_scale=hundred_scale,
            first_step_page=first_step_page,
            first_step_page_size=first_step_page_size,
            query=query,
        )



from relevanceai.constants.constants import *
from relevanceai.constants.warning import *
from relevanceai.constants.errors import *
from relevanceai.constants.messages import Messages
from relevanceai.constants.stopwords import *



import pytest
import time
from sklearn.cluster import MiniBatchKMeans

from relevanceai import Client
from relevanceai import ClusterOps
from relevanceai.dataset import Dataset

from tests.globals.constants import NOT_IMPLEMENTED


class TestClusterOps:
    vector_field = "sample_1_vector_"

    def test_dataset_cluster_1(self, test_dataset: Dataset):
        alias = "cluster_test_1"
        test_dataset.cluster(
            model="kmeans",
            cluster_config=dict(n_clusters=3),
            alias=alias,
            vector_fields=[self.vector_field],
            ingest_in_background=False,
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    def test_dataset_cluster_2(self, test_dataset: Dataset):
        alias = "cluster_test_2"
        test_dataset.cluster(
            model="kmeans",
            cluster_config=dict(n_clusters=10),
            alias=alias,
            vector_fields=[self.vector_field],
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    def test_dataset_cluster_3(self, test_dataset: Dataset):
        alias = "cluster_test_3"
        test_dataset.cluster(
            model="optics",
            alias=alias,
            vector_fields=[self.vector_field],
            include_cluster_report=False,
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    @pytest.mark.skip(reason="minibatch kmeans not implemented yet")
    def test_dataset_cluster_4(self, test_dataset: Dataset):
        alias = "cluster_test_4"
        test_dataset.cluster(
            model=MiniBatchKMeans(n_clusters=10),
            alias=alias,
            vector_fields=[self.vector_field],
        )
        assert f"_cluster_.{self.vector_field}.{alias}" in test_dataset.schema

    @pytest.mark.skip(reason="broken from refactor")
    def testClusterUtils(self, test_client: Client, test_dataset: Dataset):
        vector_field = "sample_1_vector_"
        alias = "kmeans-10"

        operator = test_dataset.cluster(
            vector_fields=[vector_field],
            model="kmeans",
            model_kwargs=dict(n_clusters=10),
            alias=alias,
        )
        operator(test_dataset)

        schema = test_dataset.schema
        assert f"_cluster_" in schema
        assert f"_cluster_.{vector_field}" in schema
        assert f"_cluster_.{vector_field}.{alias}" in schema

    def test_list_closest(self, test_client: Client, test_dataset: Dataset):
        n_clusters = 10

        clusterer = test_dataset.cluster(
            alias="new_clustering",
            model=MiniBatchKMeans(n_clusters=n_clusters),
            vector_fields=["sample_1_vector_"],
        )
        cluster_ids = ["cluster_0", "cluster_6", "cluster_3"]
        time.sleep(5)
        closests = clusterer.list_closest(
            cluster_ids=cluster_ids,
            approx=0,
            page_size=10,
            include_vector=False,
            include_count=False,
            similarity_metric="l1",
            select_fields=["_id"],
        )["results"]

        for id in cluster_ids:
            assert id in closests

        # clusterer = test_client.ClusterOps(
        #     alias="new_clustering_2",
        #     model="kmeans",
        #     n_clusters=n_clusters,
        #     vector_fields=["sample_2_vector_"],
        # )
        # clusterer.run(test_dataset)
        # closests = clusterer.list_closest(
        #     centroid_vector_fields=["sample_2_vector_"],
        #     similarity_metric="cosine",
        #     select_fields=["_id"],
        #     include_vector=True,
        # )["results"]
        # assert len(closests) == n_clusters

    @pytest.mark.skip(NOT_IMPLEMENTED)
    def test_list_furthest(self, test_client: Client, test_dataset: Dataset):
        assert False

    @pytest.mark.xfail()
    def test_merge(self, test_client: Client, test_dataset: Dataset):
        # TODO: fix this

        ALIAS = "new_merge_clustering"
        test_dataset.cluster(
            model="kmeans",
            model_kwargs={"n_clusters": 3},
            alias=ALIAS,
            vector_fields=["sample_1_vector_"],
        )

        centroids = test_client.datasets.cluster.centroids.list(
            dataset_id=test_dataset.dataset_id,
            alias=ALIAS,
            vector_fields=["sample_1_vector_"],
        )["results"]
        assert len(centroids) == 3

        ops = ClusterOps.from_dataset(
            dataset=test_dataset,
            alias=ALIAS,
            vector_fields=["sample_1_vector_"],
        )

        ops.merge(cluster_labels=["cluster_0", "cluster_1"], alias="new_clustering")

        centroids = test_client.datasets.cluster.centroids.list(
            dataset_id=test_dataset.dataset_id,
            alias=ALIAS,
            vector_fields=["sample_1_vector_"],
        )["results"]
        assert len(centroids) == 2



from relevanceai.operations.text.sentiment.sentiments import SentimentOps
from relevanceai.workflow.base import Workflow
from typing import Optional


class SentimentWorkflow(Workflow, SentimentOps):
    """
    Sentiment workflow
    """

    def __init__(
        self, model_name, workflow_alias: str = "sentiment", **workflow_kwargs
    ):
        self.model_name = model_name
        super().__init__(
            func=self.analyze_sentiment,
            workflow_alias=workflow_alias,
            **workflow_kwargs
        )

    def fit_dataset(
        self,
        dataset,
        input_field: str,
        output_field: str = "_sentiment_",
        log_to_file: bool = True,
        chunksize: int = 20,
        workflow_alias: str = "sentiment",
        notes=None,
        refresh: bool = False,
        highlight: bool = False,
        positive_sentiment_name: str = "positive",
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
    ):
        def analyze_sentiment(text):
            return self.analyze_sentiment(
                text,
                highlight=highlight,
                positive_sentiment_name=positive_sentiment_name,
                max_number_of_shap_documents=max_number_of_shap_documents,
                min_abs_score=min_abs_score,
            )

        workflow = Workflow(
            analyze_sentiment, workflow_alias=workflow_alias, notes=notes
        )
        return workflow.fit_dataset(
            dataset=dataset,
            input_field=input_field,
            output_field=output_field,
            log_to_file=log_to_file,
            chunksize=chunksize,
            refresh=refresh,
        )



from typing import List, Dict, Any, Optional, Union

from relevanceai.operations_new.run import OperationRun
from relevanceai.operations_new.scaling.models.base import ScalerModelBase


class ScalerBase(OperationRun):

    model: ScalerModelBase
    fields: List[str]
    alias: str

    def __init__(
        self,
        vector_fields: List[str],
        model: Union[str, ScalerModelBase],
        alias: Optional[str] = None,
        model_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        self.vector_fields = vector_fields

        # TODO: Add a get ailas method
        self.alias = alias  # type: ignore

        if model_kwargs is None:
            model_kwargs = {}

        self.model = self._get_model(
            model=model,
            alias=alias,
            model_kwargs=model_kwargs,
        )
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def name(self):
        return "scaing"

    def _get_model(
        self,
        model: Union[str, ScalerModelBase],
        alias: Optional[str] = None,
        model_kwargs: Optional[dict] = None,
    ) -> ScalerModelBase:

        from relevanceai.operations_new.scaling.models.sklearn import SKLearnScaler

        mapped_model = SKLearnScaler(
            model=model,
            alias=alias,
            model_kwargs=model_kwargs,
        )

        return mapped_model

    def transform(
        self,
        documents: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:

        scaled_vector_names = []

        for vector_field in self.vector_fields:
            vectors = self.get_field_across_documents(
                field=vector_field, docs=documents
            )

            reduced_vectors = self.model.fit_transform(vectors)
            scaled_vector_name = self.model.vector_name(vector_field)

            if scaled_vector_name in self.vector_fields:
                raise ValueError(
                    "Alias is already being used, Please set a different alias"
                )

            scaled_vector_names.append(scaled_vector_name)

            self.set_field_across_documents(
                field=scaled_vector_name,
                values=reduced_vectors,
                docs=documents,
            )

        # removes unnecessary info for updated_where
        updated_documents = [
            {
                key: value
                for key, value in document.items()
                if key in ["_id"] + scaled_vector_names
            }
            for document in documents
        ]

        return updated_documents
